import logging
import os
import os.path
import pickle
import time
from platform import python_version
from unittest import mock

import csp.impl.error_handling
import pytest
from fastapi.testclient import TestClient
from packaging import version

from csp_gateway import (
    Filter,
    Gateway,
    GatewaySettings,
    MountAPIKeyMiddleware,
    MountControls,
    MountFieldRestRoutes,
    MountOutputsFolder,
    MountPerspectiveTables,
    MountRestRoutes,
    MountWebSocketRoutes,
    Query,
    __version__,
)
from csp_gateway.client import GatewayClient, GatewayClientConfig
from csp_gateway.server.demo import (
    ExampleEnum,
    ExampleGatewayChannels,
    ExampleModule,
    ExampleModuleCustomTable,
)

csp.impl.error_handling.set_print_full_exception_stack(True)

if version.parse(python_version()) >= version.parse("3.11"):
    HTTPX_PATCH = "csp_gateway.client.client"
else:
    HTTPX_PATCH = "csp_gateway.client"


@pytest.fixture(scope="class")
def gateway(free_port):
    # instantiate gateway
    gateway = Gateway(
        modules=[
            ExampleModule(),
            ExampleModuleCustomTable(),
            MountControls(),
            MountOutputsFolder(),
            MountPerspectiveTables(perspective_field="perspective", layouts={"example": "test"}),
            MountRestRoutes(force_mount_all=True),
            MountFieldRestRoutes(fields=[ExampleGatewayChannels.metadata]),
            MountWebSocketRoutes(),
            MountAPIKeyMiddleware(),
        ],
        channels=ExampleGatewayChannels(),
        settings=GatewaySettings(PORT=free_port, AUTHENTICATE=True, API_KEY="test"),
    )
    return gateway


@pytest.fixture(scope="class")
def webserver(gateway):
    gateway.start(rest=True, _in_test=True)
    yield gateway
    gateway.stop()


@pytest.fixture(scope="class")
def rest_client(webserver) -> TestClient:
    return TestClient(webserver.web_app.get_fastapi())


class TestGatewayWebserver:
    server_data_flowing = None

    ############################
    # "Built in" Functionality #
    def test_openapi(self, rest_client: TestClient):
        response = rest_client.get("/openapi.json?token=test")
        assert response.status_code == 200
        json = response.json()
        assert json["info"]["title"] == "Gateway"
        assert json["info"]["version"] == __version__

    def test_docs(self, rest_client: TestClient):
        response = rest_client.get("/docs?token=test")
        assert response.status_code == 200

    def test_redoc(self, rest_client: TestClient):
        response = rest_client.get("/redoc?token=test")
        assert response.status_code == 200

    def test_unknown_404(self, rest_client: TestClient):
        response = rest_client.get("/an/unknown/route?token=test")
        assert response.status_code == 404
        assert response.json() == {"detail": "Not Found"}

    def test_static_cachecontrol(self, rest_client: TestClient):
        response = rest_client.get("/static/favicon.png")
        assert response.status_code == 200
        assert response.headers["Cache-Control"] == "public, max-age=604800"

    ######################
    # Core Functionality #
    def test_log_viewer(self, rest_client: TestClient):
        # make a temporary path that is known
        os.makedirs(os.path.join(os.getcwd(), "outputs", "testing", "temp"), exist_ok=True)
        with open(os.path.join(os.getcwd(), "outputs", "testing", "temp", "tempfile.txt"), "w") as fp:
            fp.write("test content")

        for sub_route in ("/outputs", "/outputs/testing", "/outputs/testing/temp"):
            response = rest_client.get(f"{sub_route}?token=test")
            assert response.status_code == 200

        response = rest_client.get("/outputs/somethingelseentirely?token=test")
        assert response.status_code == 404

        response = rest_client.get("/outputs/testing/temp/tempfile.txt?token=test")
        assert response.status_code == 200
        assert response.headers["content-type"].startswith("text/plain")

    def test_control_heartbeat(self, rest_client: TestClient):
        response = rest_client.get("/api/v1/controls/heartbeat?token=test")
        assert response.status_code == 200
        data = response.json()
        assert data[0]["name"] == "heartbeat"
        assert data[0]["status"] == "ok"

    #####################
    # CSP Functionality #
    def _wait_for_data(self, rest_client: TestClient):
        if self.server_data_flowing:
            return self.server_data_flowing

        # Helper function to wait for csp data to start
        # flowing before making subsequent tests
        tries = 0
        data = []
        while len(data) == 0 and tries < 50:
            # wait for some data to flow
            time.sleep(0.1)

            response = rest_client.get("/api/v1/last/example?token=test")
            assert response.status_code == 200

            data = response.json()
            assert isinstance(data, list)

            if data:
                self.server_data_flowing = data
                return data
            tries += 1
        assert "No data returned" and False

    @pytest.mark.parametrize("route", ["example", "example_list"])
    def test_csp_last(self, rest_client: TestClient, route):
        self._wait_for_data(rest_client=rest_client)

        response = rest_client.get(f"/api/v1/last/{route}?token=test")
        assert response.status_code == 200

        data = response.json()
        assert isinstance(data, list)

        datum = data[0]
        assert "x" in datum
        assert "y" in datum
        assert "data" in datum
        assert "mapping" in datum
        assert str(datum["x"]) * 3 == datum["y"]
        assert isinstance(datum["data"], list)
        assert isinstance(datum["mapping"], dict)

    def test_csp_last_basket(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)

        response = rest_client.get("/api/v1/last/basket?token=test")
        assert response.status_code == 200

        data = response.json()
        assert isinstance(data, list)
        assert len(data) == 3

        for channel_name, multiplier in ExampleEnum.__metadata__.items():
            response = rest_client.get(f"/api/v1/last/basket/{channel_name}?token=test")
            assert response.status_code == 200

            data = response.json()
            assert isinstance(data, list)
            assert len(data) == 1

            datum = data[0]
            assert "x" in datum
            assert "y" in datum
            assert str(datum["x"]) * multiplier == datum["y"]

    def test_csp_last_str_basket(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)

        response = rest_client.get("/api/v1/last/str_basket?token=test")
        assert response.status_code == 200

        data = response.json()
        assert isinstance(data, list)
        assert len(data) == 3

        for channel_name, multiplier in [("a", 1), ("b", 2), ("c", 3)]:
            response = rest_client.get(f"/api/v1/last/str_basket/{channel_name}?token=test")
            assert response.status_code == 200

            data = response.json()
            assert isinstance(data, list)
            assert len(data) == 1

            datum = data[0]
            assert "x" in datum
            assert "y" in datum
            assert str(datum["x"]) * multiplier == datum["y"]

    @pytest.mark.parametrize("route", ["example", "example_list"])
    def test_csp_next(self, rest_client: TestClient, route):
        self._wait_for_data(rest_client=rest_client)

        response = rest_client.get(f"/api/v1/next/{route}?token=test")
        assert response.status_code == 200

        data = response.json()
        assert isinstance(data, list)

        datum = data[0]
        assert "x" in datum
        assert "y" in datum
        assert str(datum["x"]) * 3 == datum["y"]

    def test_csp_next_basket(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)

        response = rest_client.get("/api/v1/next/basket?token=test")
        assert response.status_code == 200

        data = response.json()
        assert isinstance(data, list)
        assert len(data) == 3

        for channel_name, multiplier in ExampleEnum.__metadata__.items():
            response = rest_client.get(f"/api/v1/next/basket/{channel_name}?token=test")
            assert response.status_code == 200

            data = response.json()
            assert isinstance(data, list)
            assert len(data) == 1

            datum = data[0]
            assert "x" in datum
            assert "y" in datum
            assert str(datum["x"]) * multiplier == datum["y"]

    def test_csp_state(self, rest_client: TestClient):
        last_data = self._wait_for_data(rest_client=rest_client)

        # Get state data
        response = rest_client.get("/api/v1/state/example?token=test")
        assert response.status_code == 200

        state_data = response.json()
        assert isinstance(state_data, list)

        assert last_data[0] in state_data

    def test_csp_state_query(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)

        # Get state data
        response = rest_client.get('/api/v1/state/example?token=test&query={"filters":[{"attr":"x","by":{"value":1,"where":"=="}}]}')
        assert response.status_code == 200

        state_data = response.json()
        assert isinstance(state_data, list)
        assert len(state_data) == 1
        assert state_data[0]["x"] == 1

    @pytest.mark.parametrize("send_as_list", (True, False))
    def test_csp_send_validation_fails(self, rest_client: TestClient, send_as_list):
        send_data = {"x": -11, "y": "999999999"}

        response = rest_client.post("/api/v1/send/example?token=test", json=[send_data] if send_as_list else send_data)
        assert response.status_code == 422

        response_detail = response.json()["detail"]
        # This target error comes from the ExampleData
        # validation function
        target_error = "value must be non-negative."
        for val in response_detail:
            # if any message includes the target error
            # test passes
            if target_error in val["msg"]:
                return

        # should never be hit
        assert False

    @pytest.mark.parametrize("send_as_list", (True, False))
    def test_csp_send(self, rest_client: TestClient, send_as_list, caplog):
        send_data = {"x": 999, "y": "999999999", "internal_csp_struct": {"z": 15}}

        response = rest_client.post("/api/v1/send/example?token=test", json=[send_data] if send_as_list else send_data)
        assert response.status_code == 200

        return_data = response.json()
        assert isinstance(return_data, list)
        return_datum = return_data[0]
        assert "id" in return_datum
        assert (return_datum["x"], return_datum["y"], return_datum["internal_csp_struct"]) == (
            send_data["x"],
            send_data["y"],
            send_data["internal_csp_struct"],
        )

        # Perform a lookup just to make sure
        lookup_response = rest_client.get(f"/api/v1/lookup/example/{return_datum['id']}?token=test")
        assert return_data == lookup_response.json()

        time.sleep(1)
        for record in caplog.records:
            if record.levelname == "ERROR":
                raise ValueError(str(record))

    @pytest.mark.parametrize("send_as_list", (True, False))
    def test_csp_send_basket(self, rest_client: TestClient, send_as_list):
        send_data = {"x": 999, "y": "999999999"}

        response = rest_client.post("/api/v1/send/basket/A?token=test", json=[send_data] if send_as_list else send_data)
        assert response.status_code == 200

        return_data = response.json()
        assert isinstance(return_data, list)
        return_datum = return_data[0]
        assert "id" in return_datum
        assert (return_datum["x"], return_datum["y"]) == (
            send_data["x"],
            send_data["y"],
        )

    def test_csp_send_basket_whole(self, rest_client: TestClient):
        send_data = {"A": {"x": 999, "y": "999999999"}}

        response = rest_client.post("/api/v1/send/basket?token=test", json=send_data)
        assert response.status_code == 200

        return_data = response.json()
        assert isinstance(return_data, list)
        return_datum = return_data[0]
        assert (return_datum["x"], return_datum["y"]) == (
            return_datum["x"],
            return_datum["y"],
        )

    def test_csp_lookup(self, rest_client: TestClient):
        # get an existing object to fetch its ID
        data = self._wait_for_data(rest_client=rest_client)
        datum = data[0]
        assert "id" in datum
        id = datum["id"]

        # now lookup the data
        response = rest_client.get(f"/api/v1/lookup/example/{id}?token=test")
        assert data == response.json()

    def test_csp_lookup_list(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)
        # get an existing object to fetch its ID
        response = rest_client.get("/api/v1/last/example_list?token=test")
        data = response.json()
        datum = data[0]
        assert "id" in datum
        id = datum["id"]

        # now lookup the data
        response = rest_client.get(f"/api/v1/lookup/example_list/{id}?token=test")
        assert datum == response.json()[0]

    def test_csp_toplevel_last(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)

        response_last = rest_client.get("/api/v1/last?token=test")
        assert response_last.status_code == 200
        assert sorted(response_last.json()) == [
            "basket",
            "controls",
            "example",
            "example_list",
            "never_ticks",
            "str_basket",
        ]

    def test_csp_toplevel_next(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)

        response_last = rest_client.get("/api/v1/next?token=test")
        assert response_last.status_code == 200
        assert sorted(response_last.json()) == [
            "basket",
            "controls",
            "example",
            "example_list",
            "never_ticks",
            "str_basket",
        ]

    def test_csp_toplevel_state(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)
        response_state = rest_client.get("/api/v1/state?token=test")
        assert response_state.status_code == 200
        assert sorted(response_state.json()) == ["example"]

    def test_csp_toplevel_send(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)
        response_send = rest_client.get("/api/v1/send?token=test")
        assert response_send.status_code == 200
        assert sorted(response_send.json()) == [
            "basket",
            "basket/A",
            "basket/B",
            "basket/C",
            "controls",
            "example",
        ]

    @mock.patch(HTTPX_PATCH + ".POST", autospec=True)
    @mock.patch(HTTPX_PATCH + ".GET", autospec=True)
    def test_gateway_client(self, mock_get, mock_post, rest_client: TestClient, gateway: Gateway):
        mock_get.side_effect = rest_client.get
        mock_post.side_effect = rest_client.post

        gateway_client = GatewayClient(GatewayClientConfig(port=gateway.settings.PORT, authenticate=True, api_key="test"))
        self._wait_for_data(rest_client=rest_client)
        response_state = gateway_client.state()
        assert sorted(list(gateway_client.openapi_spec.keys())) == ["components", "info", "openapi", "paths"]
        assert "/api/v1/last/example" in gateway_client.openapi_spec["paths"].keys()
        assert "/api/v1/last/example_list" in gateway_client.openapi_spec["paths"].keys()
        assert response_state == ["example"]

        for route in ["example", "example_list"]:
            data = gateway_client.last(route)
            assert isinstance(data, list)

            datum = data[0]
            assert "x" in datum
            assert "y" in datum
            assert "data" in datum
            assert "mapping" in datum
            assert str(datum["x"]) * 3 == datum["y"]
            assert isinstance(datum["data"], list)
            assert isinstance(datum["mapping"], dict)

        data_response = gateway_client.last("example", return_raw_json_override=False)
        expected_columns = {"id", "timestamp", "x", "y", "data", "dt", "d", "internal_csp_struct.z", "mapping"}
        data_pd = data_response.as_pandas_df()
        actual_columns_pd = set(data_pd.columns)
        assert expected_columns.issubset(actual_columns_pd)

        data_pl = data_response.as_polars_df()
        actual_columns_pl = set(data_pl.columns)
        assert expected_columns.issubset(actual_columns_pl)

        send_data = {"x": 999, "y": "999999999"}
        return_data = gateway_client.send("basket/A", send_data)
        assert isinstance(return_data, list)
        return_datum = return_data[0]
        assert "id" in return_datum
        assert (return_datum["x"], return_datum["y"]) == (
            send_data["x"],
            send_data["y"],
        )

        send_data = {"x": 9999, "y": "9999999999"}
        return_data = gateway_client.send("basket/A", [send_data])
        assert isinstance(return_data, list)
        return_datum = return_data[0]
        assert "id" in return_datum
        assert (return_datum["x"], return_datum["y"]) == (
            send_data["x"],
            send_data["y"],
        )

        return_data = gateway_client.state("example", query=Query(filters=[Filter(attr="x", by={"value": 1, "where": "=="})]))
        assert isinstance(return_data, list)
        return_datum = return_data[0]
        assert "id" in return_datum
        assert return_datum["x"] == 1

    def test_csp_specific_channels(self, caplog):
        caplog.set_level(logging.INFO)
        gateway = Gateway(
            modules=[
                ExampleModule(),
                ExampleModuleCustomTable(),
                MountControls(),
                MountOutputsFolder(),
                MountPerspectiveTables(perspective_field="perspective", layouts={"example": "test"}),
                MountRestRoutes(
                    mount_send=[
                        ExampleGatewayChannels.example,
                        ExampleGatewayChannels.never_ticks,
                    ],
                    mount_last=[
                        ExampleGatewayChannels.example,
                        ExampleGatewayChannels.basket,
                    ],
                ),
                MountFieldRestRoutes(fields=[ExampleGatewayChannels.metadata]),
                MountWebSocketRoutes(),
            ],
            channels=ExampleGatewayChannels(),
        )
        gateway.start(rest=True, _in_test=True)
        rest_client = TestClient(gateway.web_app.get_fastapi())
        self._wait_for_data(rest_client=rest_client)

        response_last = rest_client.get("/api/v1/last?token=test")
        assert response_last.status_code == 200
        assert sorted(response_last.json()) == ["basket", "example"]

        response_send = rest_client.get("/api/v1/send?token=test")
        assert response_send.status_code == 200
        assert sorted(response_send.json()) == [
            "example",
        ]

        send_data = {"A": {"x": 999, "y": "999999999"}}

        response = rest_client.post("/api/v1/send/basket?token=test", json=send_data)
        assert response.status_code == 404
        gateway.stop()
        assert f"Requested channels missing send routes are: ['{ExampleGatewayChannels.never_ticks}']" in caplog.text

    def test_websocket_subscribe_unsubscribe(self, rest_client: TestClient):
        self._wait_for_data(rest_client=rest_client)

        with rest_client.websocket_connect("/api/v1/stream?token=test") as websocket:
            # subscribe
            websocket.send_json({"action": "subscribe", "channel": "example"}, mode="text")

            # receive data
            data = websocket.receive_json()
            assert "channel" in data
            assert "data" in data

            msg = data["data"][0]
            assert "id" in msg
            assert "timestamp" in msg
            assert "x" in msg
            assert "y" in msg

            # send data
            websocket.send_json(
                {
                    "action": "send",
                    "channel": "example",
                    "data": {"x": 12345, "y": "54321"},
                }
            )

            data = websocket.receive_json()

            assert "channel" in data
            assert "data" in data

            msg = data["data"][0]
            assert "id" in msg
            assert "timestamp" in msg
            assert msg["x"] == 12345
            assert msg["y"] == "54321"

            # send data as list
            websocket.send_json(
                {
                    "action": "send",
                    "channel": "example",
                    "data": [{"x": 54321, "y": "12345"}],
                }
            )

            data = websocket.receive_json()
            assert "channel" in data
            assert "data" in data

            msg = data["data"][0]
            assert "id" in msg
            assert "timestamp" in msg
            assert msg["x"] == 54321
            assert msg["y"] == "12345"

            # unsubscribe
            websocket.send_json({"action": "unsubscribe", "channel": "example"}, mode="text")
            # subscribe dict-basket
            # We subscribe via enum name
            websocket.send_json({"action": "subscribe", "channel": "basket", "key": "A"}, mode="text")

            data = websocket.receive_json()

            assert data["channel"] == "basket"
            assert data["key"] == "A"
            assert "data" in data

            # send data as list
            websocket.send_json(
                {
                    "action": "send",
                    "channel": "basket",
                    "key": "A",
                    "data": [{"x": 9989, "y": "1273"}],
                }
            )
            data = websocket.receive_json()

            assert data["channel"] == "basket"
            assert data["key"] == "A"
            msg = data["data"][0]

            assert "id" in msg
            assert "timestamp" in msg
            assert msg["x"] == 9989
            assert msg["y"] == "1273"

            # send data as single item
            # but we don't receive it back
            # since we aren't subscribed.
            websocket.send_json(
                {
                    "action": "send",
                    "channel": "basket",
                    "key": "B",
                    "data": {"x": 9989, "y": "1273"},
                }
            )
            data = websocket.receive_json()
            assert data["channel"] == "basket"
            # We sent to B
            assert data["key"] == "A"
            msg = data["data"][0]
            assert "id" in msg
            assert "timestamp" in msg

            websocket.send_json({"action": "subscribe", "channel": "str_basket", "key": "a"}, mode="text")

            # unsubscribe dict-basket
            # since no key, we unsubscribe from everything
            websocket.send_json({"action": "unsubscribe", "channel": "basket"}, mode="text")

            data = websocket.receive_json()

            # We unsubscribed from 'basket'
            assert data["channel"] == "str_basket"
            assert data["key"] == "a"
            msg = data["data"][0]
            assert "id" in msg
            assert "timestamp" in msg

            # unsubscribe dict-basket
            # since no key, we unsubscribe from everything
            websocket.send_json({"action": "unsubscribe", "channel": "str_basket", "key": "a"}, mode="text")

            websocket.send_json(
                {
                    "action": "send",
                    "channel": "example",
                    "data": {"x": 12345, "y": "54321"},
                }
            )
            with pytest.raises(Exception):
                websocket._send_queue.get(timeout=2.0)

    def test_perspective_tables(self, rest_client: TestClient):
        response_last = rest_client.get("/api/v1/perspective/tables?token=test")
        assert response_last.status_code == 200
        assert sorted(response_last.json().keys()) == [
            "basket",
            "controls",
            "example",
            "example_list",
            "my_custom_table",
            # "never_ticks",  # NOTE: never_ticks has no data, so table is omitted
            "str_basket",
        ]

    def test_perspective_layouts(self, rest_client: TestClient):
        response_last = rest_client.get("/api/v1/perspective/layouts?token=test")
        assert response_last.status_code == 200
        assert response_last.json() == {"example": "test"}

    def test_fields(self, rest_client: TestClient):
        response_field = rest_client.get("/api/v1/field/metadata?token=test")
        assert response_field.status_code == 200
        assert response_field.json() == {"name": "Demo"}

        response_field = rest_client.get("/api/v1/field/garbage?token=test")
        assert response_field.status_code == 404

        # Get list of fields
        response = rest_client.get("/api/v1/field?token=test")
        assert response.status_code == 200

        assert response.json() == ["metadata"]

    def test_stream(self, rest_client: TestClient):
        response_stream = rest_client.get("/api/v1/stream?token=test")
        expected = [
            "basket/A",
            "basket/B",
            "basket/C",
            "controls",
            "example",
            "example_list",
            "heartbeat",
            "never_ticks",
            "str_basket/a",
            "str_basket/b",
            "str_basket/c",
        ]
        assert response_stream.status_code == 200
        assert sorted(response_stream.json()) == expected


def test_MountRestRoutes_validator(caplog):
    """Test for backwards compatibility given API change of mount_all->force_mount_all"""
    r = MountRestRoutes(mount_all=False)
    assert not r.force_mount_all
    assert "mount_all is deprecated, please use force_mount_all instead" in caplog.text

    r = MountRestRoutes(mount_all=True)
    assert r.force_mount_all


def test_MountPerspectiveTables_pickleable():
    """Test that MountPerspectiveTables is pickleable"""
    mpt = MountPerspectiveTables()
    pickle.dumps(mpt)



import logging
from collections import defaultdict, deque
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional, Set, Tuple, TypeVar, Union

import csp
import numpy as np
from ccflow import BaseModel
from ccflow.serialization import make_ndarray_orjson_valid
from csp import Struct, ts
from csp.impl.enum import Enum, EnumMeta
from csp.impl.types.tstype import isTsType
from pydantic import Field, PrivateAttr, TypeAdapter

from csp_gateway.server.gateway.csp import ChannelsType
from csp_gateway.server.gateway.csp.channels import _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD
from csp_gateway.utils import (
    get_dict_basket_value_tstype,
    is_dict_basket,
    is_list_basket,
)

__all__ = (
    "JSONConverter",
    "EncodedEngineCycle",
    "read_engine_encoding_json_with_duckdb",
)

_KEY_TYPE = Union[str, Enum]
T = TypeVar("T")
K = TypeVar("K")

_TIMEDELTA_TYPE_ADAPTER = TypeAdapter(timedelta)


class ChannelValueModel(BaseModel):
    """Pydantic model representing a tick on a specific channel."""

    channel: str
    value: Any
    dict_basket_key: Optional[_KEY_TYPE] = None
    timestamp: datetime


class EncodedEngineCycle(csp.Struct):
    """This is a csp struct representing a single engine cycle."""

    encoding: str
    csp_timestamp: datetime


def read_engine_encoding_json_with_duckdb(channels: ChannelsType, filename: str):
    """Helper function to read a newline-delimited json file containing engine cycle ticks into duckdb."""
    import duckdb

    channel_names = channels._snapshot_model.model_fields.keys()
    columns = {channel: "JSON" if channel != _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD else "TIMESTAMP" for channel in channel_names}
    return duckdb.read_json(filename, columns=columns, format="newline_delimited")


def _convert_orjson_compatible(obj: Any):
    if isinstance(obj, Struct):
        return {
            _convert_orjson_compatible(k): _convert_orjson_compatible(getattr(obj, k))
            for k in obj.__full_metadata__
            if (not k.startswith("_")) and hasattr(obj, k)
        }
    if isinstance(obj, dict):
        return {_convert_orjson_compatible(k): _convert_orjson_compatible(v) for k, v in obj.items()}
    if isinstance(obj, (list, set, tuple)):
        return [_convert_orjson_compatible(val) for val in obj]
    if isinstance(obj, (Enum, EnumMeta)):
        return obj.name
    if isinstance(obj, timedelta):
        return _TIMEDELTA_TYPE_ADAPTER.dump_python(obj, mode="json")
    if isinstance(obj, np.ndarray):
        return make_ndarray_orjson_valid(obj)
    return obj


def _create_snapshot_dict(all_data: List[ChannelValueModel]) -> Dict[str, Any]:
    res = defaultdict(dict)
    res[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD] = all_data[0].timestamp
    for data in all_data:
        channel = data.channel
        value = _convert_orjson_compatible(data.value)
        if key := data.dict_basket_key:
            res[channel][_convert_orjson_compatible(key)] = value
        else:
            res[channel] = value
        res[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD] = min(res.get(_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD), data.timestamp)
    return res


@csp.node
def _create_tuple_not_dict_basket(
    channel: str,
    value: ts[object],
) -> ts[tuple]:
    return (channel, (value,))


@csp.node
def _create_tuple_dict_basket(
    channel: str,
    value: ts[object],
    dict_basket_key: _KEY_TYPE,
) -> ts[tuple]:
    return (channel, (dict_basket_key, value))


@csp.node
def _deserialize_snapshot_dict(encoded_snapshot: ts[dict], typ: "T", log_lagging_engine_cycles: bool) -> ts["T"]:
    with csp.alarms():
        # Define an alarm time-series of type bool
        alarm = csp.alarm(bool)

    with csp.start():
        stack = deque()

    if csp.ticked(encoded_snapshot):
        snapshot = typ.model_validate(encoded_snapshot)
        timestamp = getattr(snapshot, _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD).replace(tzinfo=None)
        csp_now = csp.now()
        if timestamp > csp_now:
            csp.schedule_alarm(alarm, timestamp, True)
            stack.append(snapshot)
        else:
            if timestamp < csp_now and log_lagging_engine_cycles:
                logging.info(
                    f"Timestamp for a replayed engine cycle is: {timestamp} "
                    + f"which is behind csp engine time: {csp_now}. Engine cycle is: {snapshot}"
                )
            return snapshot

    if csp.ticked(alarm):
        return stack.popleft()


@csp.node
def _deserialize_snapshot_str(encoded_snapshot: ts[str], typ: "T", log_lagging_engine_cycles: bool) -> ts["T"]:
    with csp.alarms():
        # Define an alarm time-series of type bool
        alarm = csp.alarm(bool)

    with csp.start():
        stack = deque()

    if csp.ticked(encoded_snapshot):
        snapshot = typ.model_validate_json(encoded_snapshot)
        timestamp = getattr(snapshot, _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD).replace(tzinfo=None)
        csp_now = csp.now()
        if timestamp > csp_now:
            csp.schedule_alarm(alarm, timestamp, True)
            stack.append(snapshot)
        else:
            if timestamp < csp_now and log_lagging_engine_cycles:
                logging.info(
                    f"Timestamp for a replayed engine cycle is: {timestamp} "
                    + f"which is behind csp engine time: {csp_now}. Engine cycle is: {snapshot}"
                )
            return snapshot

    if csp.ticked(alarm):
        return stack.popleft()


def _deserialize_snapshot(
    encoded_snapshot: Union[ts[str], ts[Dict[str, Any]]],
    typ: Any,
    log_lagging_engine_cycles: bool = False,
):
    if encoded_snapshot.tstype.typ is str:
        return _deserialize_snapshot_str(encoded_snapshot, typ, log_lagging_engine_cycles)
    return _deserialize_snapshot_dict(encoded_snapshot, typ, log_lagging_engine_cycles)


class JSONConverter(BaseModel):
    """
    This class is instantiated within a Gateway Module to wire the channel for
    json encoding/decoding. The encode/decode functions are called to perform the
    respective wirings. Only channels in the snapshot model of the channels class
    are included.
    """

    decode_channels: List[str]
    encode_channels: List[str]
    channels: ChannelsType
    flag_updates: Dict[str, List[Tuple[str, bool]]] = Field(
        default_factory=dict,
        description="""
        A mapping of channels to a list of tuples each containing:
        A field name and the flag it should be updated to, when in Read mode.
        Only works for non-dict basket channels
        """,
    )
    decode_exclude_fields: Optional[Set[str]] = Field(
        None,
        description="""
        Set of fields to exclude when we are decoding from a source.
        Does not throw an error if the field does not exist.
        """,
    )
    log_lagging_engine_cycles: bool = Field(False, description="Whether we log when engine cycles are lagging from the current engine time.")
    _decode_channel_dict_basket_keys: Dict[str, List[_KEY_TYPE]] = PrivateAttr(default_factory=dict)
    _snapshot_type: type = PrivateAttr(default=None)

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._snapshot_type = type(self.channels)._snapshot_model

        # get the keys for every channel we are decoding
        for field in self.decode_channels:
            outer_type = self.channels.get_outer_type(field)
            if is_dict_basket(outer_type):
                edge = self.channels.get_channel(field)
                self._decode_channel_dict_basket_keys[field] = list(edge.keys())

    def get_channel_from_snapshot(
        self,
        snapshot: ts[object],
        field: _KEY_TYPE,
        value_type: type,
        keys: Optional[List[_KEY_TYPE]] = None,
    ):
        if keys:
            if updates := self.flag_updates.get(field):
                raise ValueError("Flag updates cannot be called on csp dict baskets" + f"Got flag {updates = }")
            return self._get_dict_basket(snapshot, field, keys, value_type)
        return self._get_non_dict_basket(snapshot, field, value_type)

    @csp.node
    def _get_dict_basket(
        self,
        snapshot: ts[object],
        field: _KEY_TYPE,
        keys: ["K"],
        value_type: "T",
    ) -> csp.OutputBasket(Dict["K", ts["T"]], shape="keys"):  # noqa
        if csp.ticked(snapshot):
            dict_basket = getattr(snapshot, field)
            if dict_basket is not None:
                return {k: v for k, v in dict_basket.items() if v is not None}

    @csp.node
    def _get_non_dict_basket(
        self,
        snapshot: ts[object],
        field: _KEY_TYPE,
        value_type: "T",
    ) -> ts["T"]:
        if csp.ticked(snapshot):
            pydantic_value = getattr(snapshot, field)
            if pydantic_value is not None:
                channel_flag_updates = self.flag_updates.get(field, [])
                value = pydantic_value
                # TODO: ponder how to better do this
                if self.decode_exclude_fields:
                    for field in self.decode_exclude_fields:
                        if field == "id":
                            value.id = value.generate_id()
                        elif field == "timestamp":
                            value.timestamp = datetime.now(timezone.utc)
                        elif hasattr(value, field):
                            delattr(value, field)
                if isinstance(value, (list, tuple)):
                    for attribute, updated_flag in channel_flag_updates:
                        for obj in value:
                            setattr(obj, attribute, updated_flag)
                else:
                    for attribute, updated_flag in channel_flag_updates:
                        setattr(value, attribute, updated_flag)
                return value

    def encode(self, writing_mode: ts[bool] = None) -> ts[EncodedEngineCycle]:
        """Given the channels object and the channels we are encoding, we combine
        them all into a fat pipe that is a long ts[str] object that stores all the
        ticks for every specified channel in each engine cycle.

        writing_mode: ts[bool]
            This ts object is provided by the Gateway Module that instantiates the Json Handler to
            specify when to encode or decode. We want to avoid encoding channel ticks that are
            caused by a json string that we are decoding
        """
        # If we don't specify a writing mode, we will assume we write
        if writing_mode is None:
            writing_mode = csp.const(True)

        snapshot_model_fields = self._snapshot_type.model_fields

        all_data = []  # input as (field, data) where data is a ticked channel, or a tuple (key, ticked_edge) in a dict basket
        for field in self.encode_channels:
            if field not in snapshot_model_fields:
                continue

            outer_type = self.channels.get_outer_type(field)
            edge = self.channels.get_channel(field)

            # TODO: list baskets not supported yet in channels model
            if is_list_basket(outer_type):
                raise TypeError("List Baskets not yet supported in channels model")

            # We do this so that when we collect at the bottom, we get all channels that ticked together
            # at once. Different values in a dict basket can tick out of sync with each other
            elif is_dict_basket(outer_type):
                data = [_create_tuple_dict_basket(channel=field, value=csp.filter(writing_mode, v), dict_basket_key=k) for k, v in edge.items()]
                all_data.extend(data)

            elif isTsType(outer_type):
                all_data.append(_create_tuple_not_dict_basket(channel=field, value=csp.filter(writing_mode, edge)))

        return self._create_encoding(csp.collect(all_data))

    def decode(self, fat_pipe: Union[ts[str], ts[Dict[str, Any]]]) -> None:
        """
        Takes in the fat pipe with all of the encoded channel ticks,
        then demultiplexes them, deserializes the values, and
        sets the channels accordingly.
        """
        ts_snapshot_model = _deserialize_snapshot(
            fat_pipe,
            self._snapshot_type,
            log_lagging_engine_cycles=self.log_lagging_engine_cycles,
        )
        for field in self.decode_channels:
            outer_type = self.channels.get_outer_type(field)
            # having basket_keys means the channel is a dict_basket
            if basket_keys := self._decode_channel_dict_basket_keys.get(field):
                outer_type = get_dict_basket_value_tstype(outer_type)

            output = self.get_channel_from_snapshot(
                ts_snapshot_model,
                field,
                outer_type.typ,
                basket_keys,
            )
            self.channels.set_channel(field, output)

    @csp.node
    def _create_encoding(self, all_data: ts[List[tuple]]) -> ts[EncodedEngineCycle]:
        if csp.ticked(all_data):
            values = defaultdict(dict)
            for field, val in all_data:
                if len(val) == 1:
                    values[field] = val[0]
                else:
                    values[field][val[0]] = val[1]
            values[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD] = csp.now()
            res = self._snapshot_type.model_validate(values).model_dump_json()
            return EncodedEngineCycle(encoding=res + "\n", csp_timestamp=csp.now())  # we manually add the new line at the end

    def __hash__(self):
        # So that csp doesn't complain about inability to memoize
        return id(self)



import asyncio
import logging
from typing import List, Optional, Set

from fastapi import APIRouter, BackgroundTasks, HTTPException, Request

from csp_gateway.utils import Controls

from ..utils import get_default_responses
from .shared import prepare_response

log = logging.getLogger(__name__)


_WAIT_THRESHOLD = 0.1

__all__ = (
    "add_controls_routes",
    "add_controls_available_channels",
)


def add_controls_routes(api_router: APIRouter, field: str) -> None:
    if field == "heartbeat":
        # Add heartbeat channel
        @api_router.get(
            "/heartbeat",
            responses=get_default_responses(),
            response_model=Controls,
            name="Get Heartbeat",
        )
        async def heartbeat(request: Request) -> Controls:
            """
            This endpoint is a lightweight `ping`/`pong` endpoint that can be used to determine the status of the underlying webserver.
            """
            data = Controls(name="heartbeat")

            # Throw 404 if not a supported channel
            if not hasattr(request.app.gateway.channels, "controls"):
                raise HTTPException(status_code=404, detail="Channel not found: controls")

            # send data to csp
            request.app.gateway.channels.send("controls", data)

            # don't care about the result
            while data.status != "ok":
                await asyncio.sleep(_WAIT_THRESHOLD)

            return prepare_response(data, is_list_model=False)

    elif field == "stats":

        @api_router.get(
            "/stats",
            responses=get_default_responses(),
            response_model=Controls,
            name="Get CSP Stats",
        )
        async def stats(request: Request) -> Controls:
            """This endpoint will collect and return various engine and system stats, including:

            - CPU utilization (`cpu`)
            - Virtual memory utilization (`memory`)
            - Total memory available (`memory-total`)
            - Current system time (`now`)
            - CSP engine time (`csp-now`)
            - Hostname (`host`)
            - Username (`user`)
            """
            data = Controls(name="stats")

            # Throw 404 if not a supported channel
            if not hasattr(request.app.gateway.channels, "controls"):
                raise HTTPException(status_code=404, detail="Channel not found: controls")

            # send data to csp
            request.app.gateway.channels.send("controls", data)

            while not data.data:
                await asyncio.sleep(_WAIT_THRESHOLD)
            data.update_str()

            return prepare_response(data, is_list_model=False)

    elif field == "shutdown":

        @api_router.post(
            "/shutdown",
            responses=get_default_responses(),
            response_model=Controls,
            name="Shutdown Server",
        )
        async def shutdown(request: Request, background_tasks: BackgroundTasks) -> Controls:
            """
            **WARNING:** Use this endpoint with caution.

            This endpoint will cleanly shutdown the engine and webserver. It is used for the kill switch in UIs.
            """
            # FIXME ugly
            background_tasks.add_task(request.app.gateway.stop, user_initiated=True)

            data = Controls(name="shutdown", status="ok")
            return prepare_response(data, is_list_model=False)

    else:
        raise ValueError(f"Unsupported controls field: {field}. Supported fields are 'heartbeat', 'stats', and 'shutdown'.")


def add_controls_available_channels(api_router: APIRouter, fields: Optional[Set[str]] = None) -> None:
    @api_router.get(
        "/",
        responses=get_default_responses(),
        response_model=List[str],
    )
    async def get_controls(request: Request) -> List[str]:
        """
        This endpoint will return a list of string values of all available channels under the `/controls` route.
        """
        return sorted(fields if fields else ("heartbeat", "stats", "shutdown"))



import abc
from typing import Any, List, Union, get_args, get_origin

import csp
from ccflow import BaseModel
from csp import ts

from csp_gateway.utils import GatewayStruct

__all__ = ("KafkaChannelProcessor",)


class KafkaChannelProcessor(BaseModel, abc.ABC):
    """
    Process channel inputs before sending to Kafka, or before propagating into the graph.
    """

    @abc.abstractmethod
    def process(self, obj: Union[List[GatewayStruct], GatewayStruct], topic: str, key: str) -> Any:
        raise NotImplementedError()

    def apply_process(self, typ: Any, obj: ts[object], topic: str, key: str) -> ts[object]:
        """Applies a function to a ticking edge.

        Args:
            typ: The class of the processed object, according to csp.
            obj: The ticking edge that ticks values of the object
            topic: The Kafka topic
            key: The Kafka key

        Returns:
            A ticking edge with the processed output, of type specified by `typ`
        """

        actual_type = typ
        # csp doesnt like typing.List, so we replace it if we see it
        if get_origin(typ) is list:
            arg = get_args(typ)[0]
            actual_type = [arg]

        res = csp.apply(obj, lambda x, topic=topic, key=key: self.process(x, topic, key), object)
        flag = csp.apply(res, lambda x: x is not None, bool)

        filtered_res = csp.filter(flag, res)
        # safer than static_cast since csp does type checking at runtime for us
        return csp.dynamic_cast(filtered_res, actual_type)



from typing import Any, Optional

from fastapi import Depends, HTTPException, Query
from pydantic import Json, ValidationError

from .web.query import Query as QueryParamType


def json_param(param_name: str, model: Any, **query_kwargs):
    """Parse JSON-encoded query parameters as pydantic models.
    The function returns a `Depends()` instance that takes the JSON-encoded value from
    the query parameter `param_name` and converts it to a Pydantic model, defined
    by the `model` attribute.
    """

    def get_parsed_object(value: Optional[Json] = Query(default=None, alias=param_name, **query_kwargs)):
        try:
            if value is None:
                return None

            return model.model_validate(value)

        except ValidationError as err:
            raise HTTPException(400, detail=err.errors())

    return Depends(get_parsed_object)


def query_json():
    # NOTE: if we switch querys back to be csp structs, do this:
    # return json_param("query", QueryParamType.__pydantic_model__, description="Query Parameters")
    return json_param("query", QueryParamType, description="Query Parameters")



class GatewayException(Exception): ...


class NoProviderException(GatewayException): ...


class ServerException(GatewayException): ...


class ServerRouteNotMountedException(ServerException): ...


class ServerRouteNotFoundException(ServerException): ...


class ServerUnprocessableException(ServerException): ...


class ServerUnknownException(ServerException): ...


class _Controls(GatewayException):
    def __init__(self, control: str):
        super().__init__("Control: {}".format(control))



from functools import wraps
from typing import Any, Awaitable, Callable, Optional, Tuple, TypeVar

from fastapi.exceptions import HTTPException, WebSocketException
from starlette.requests import HTTPConnection
from starlette.status import WS_1008_POLICY_VIOLATION
from starlette.websockets import WebSocket


def get_authorization_scheme_param(
    authorization_header_value: Optional[str],
) -> Tuple[str, str]:
    if not authorization_header_value:
        return "", ""
    scheme, _, param = authorization_header_value.partition(" ")
    return scheme, param


_SecurityDepFunc = TypeVar("_SecurityDepFunc", bound=Callable[[Any, HTTPConnection], Awaitable[Any]])


def handle_exc_for_ws(func: _SecurityDepFunc) -> _SecurityDepFunc:
    @wraps(func)
    async def wrapper(self: Any, request: HTTPConnection) -> Any:
        try:
            return await func(self, request)
        except HTTPException as e:
            if not isinstance(request, WebSocket):
                raise e
            # close before accepted with result a HTTP 403 so the exception argument is ignored
            # ref: https://asgi.readthedocs.io/en/latest/specs/www.html#close-send-event
            raise WebSocketException(code=WS_1008_POLICY_VIOLATION, reason=e.detail) from None

    return wrapper  # type: ignore



from pydantic import TypeAdapter

from csp_gateway.server import ChannelSelection
from csp_gateway.testing.shared_helpful_classes import MyGatewayChannels


def test_selection():
    selection = ChannelSelection()
    channels = [
        "my_channel",
        "my_list_channel",
        "my_enum_basket",
        "my_str_basket",
        "my_enum_basket_list",
        "my_array_channel",
    ]
    static = ["my_static", "my_static_dict", "my_static_list"]

    assert selection.select_from(MyGatewayChannels) == channels
    assert selection.select_from(MyGatewayChannels, state_channels=True) == [
        "s_my_channel",
        "s_my_list_channel",
    ]
    assert selection.select_from(MyGatewayChannels, static_fields=True) == static


def test_selection_duplicates():
    channels = ["my_channel", "my_enum_basket"]
    static = ["my_static", "my_static_dict"]
    selection = ChannelSelection(include=channels + channels + static)

    assert selection.select_from(MyGatewayChannels) == channels
    assert selection.select_from(MyGatewayChannels, state_channels=True) == ["s_my_channel"]
    assert selection.select_from(MyGatewayChannels, static_fields=True) == static


def test_selection_all_fields():
    channels = ["my_enum_basket", "my_channel", "s_my_channel"]
    static = ["my_static", "my_static_dict"]
    selection = ChannelSelection(include=channels + static)

    assert selection.select_from(MyGatewayChannels, all_fields=True) == [
        "my_enum_basket",
        "my_channel",
        "s_my_channel",
        "my_static",
        "my_static_dict",
    ]


def test_selection_include():
    channels = ["my_channel", "my_enum_basket"]
    static = ["my_static", "my_static_dict"]
    selection = ChannelSelection(include=channels + static)

    assert selection.select_from(MyGatewayChannels) == channels
    assert selection.select_from(MyGatewayChannels, state_channels=True) == ["s_my_channel"]
    assert selection.select_from(MyGatewayChannels, static_fields=True) == static
    assert selection.select_from(MyGatewayChannels, all_fields=True) == [
        "my_channel",
        "my_enum_basket",
        "my_static",
        "my_static_dict",
    ]


def test_selection_exclude():
    selection = ChannelSelection(exclude=["my_channel", "my_enum_basket", "my_static"])
    channels = [
        "my_list_channel",
        "my_str_basket",
        "my_enum_basket_list",
        "my_array_channel",
    ]
    static = ["my_static_dict", "my_static_list"]

    assert selection.select_from(MyGatewayChannels) == channels
    assert selection.select_from(MyGatewayChannels, state_channels=True) == ["s_my_list_channel"]
    assert selection.select_from(MyGatewayChannels, static_fields=True) == static


def test_validate():
    selection = TypeAdapter(ChannelSelection).validate_python(["my_channel", "my_enum_basket"])
    assert selection.include == ["my_channel", "my_enum_basket"]

    selection = TypeAdapter(ChannelSelection).validate_python(None)
    assert selection.include is None



import json
import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, TypeVar

import csp
from csp import ts
from pydantic import BaseModel

from csp_gateway import GatewayChannels, GatewayModule

if TYPE_CHECKING:
    from sqlalchemy.engine.base import Engine

logger = logging.getLogger(__name__)


T = TypeVar("T")


class SQLACnxDetails(BaseModel):
    username: str = ""
    password: str = ""
    host: str = ""
    db: str = ""
    port: int = 5432
    engine: str = "postgresql"  # TODO: this should be enum probably
    driver: str = "ODBC Driver 17 for SQL Server"  # pyodbc.drivers()[-1]

    def get_cnx_string(self) -> str:
        if self.engine == "mssql+pyodbc":  # requires the odbc str and driver
            return f"{self.engine}://{self.host}/{self.db}?driver={self.driver.replace(' ', '+')}&trusted_connection=yes"
        elif self.engine == "sqlite":
            return f"{self.engine}:///{self.host}"
        else:
            return f"{self.engine}://{self.username}:{self.password}@{self.host}:{self.port}/{self.db}"


class ChannelSchemaConfig(BaseModel):
    """
    channel_name: str
        name of channel to check for ticks
    table: str
        name of table to write to
    fields: List[str] = []
        specific of list of fields to check the channel underlying attribute for
        if empty, will use all known attributes from object
    rename_fields: Dict[str, str] = {}
        rename any fields from the attribute to the table name ( e.x. { 'attr_field': 'table_field', ...} )
    augmentation_fields: Dict[str, str] = {}
        add additional details to the insert query (e.x. {'trader_id': 'ZZZZ'})
    """

    channel_name: str  # name of channel to write
    table: str
    fields: List[str] = []  # if empty will just utilize all fields
    rename_fields: Dict[str, str] = {}
    augmentation_fields: Dict[str, str] = {}


class PublishSQLA(GatewayModule):
    """
    The PublishSQLA module allows for formatted writing of items to tables.

    Note: requires sqlalchemy

    ...

    Kwargs
    ------
    cnx_details: SQLACnxDetails
        the struct detailing how to create the connection string
    schema_configs: List[ChannelSchemaConfig]
        configuration for each channel to write to tables
    n_tries: int (default = 1)
        how many times should it try to write to the db before giving up
    fail_after_retry: bool (default=False)
        if we pass n_tries without success, should we raise an error in the node


    Example
    -------
    ```python
    from csp_gateway.server.modules.sql import SQLACnxDetails, ChannelSchemaConfig, PublishSQLA
    cnx_detail = SQLACnxDetails(
        engine='mssql+pydobc',
        host='MYSQLCONNECTIONENGINE',
        db='MyDatabase'
    )
    schema_configs = [
        ChannelSchemaConfig(
            channel_name='my_channel',
            table='my_channels_table',
            fields=['a','b','c'],
            rename_fields={'a':'not_a'},
            augmentation_fields= {'trader_id': 'ZZZZ'} # add this to every record
        )
    ]
    modules.append(
        PublishSQLA(cnx_details=cnx_details, schema_configs=schema_configs)
    )
    ```

    """

    cnx_details: SQLACnxDetails
    schema_configs: List[ChannelSchemaConfig] = []
    n_tries: int = 1
    fail_after_retry: bool = False

    # not to be set
    channel_map: Dict[str, ChannelSchemaConfig] = {}
    engine: "Engine" = None

    def __init__(self, *args, **kwargs):
        from sqlalchemy.engine.base import Engine  # noqa: F401

        PublishSQLA.model_rebuild()
        super().__init__(*args, **kwargs)

    def connect(self, channels: GatewayChannels):
        from sqlalchemy import create_engine

        self.engine = create_engine(self.cnx_details.get_cnx_string(), pool_pre_ping=True)
        for config in self.schema_configs:
            self.channel_map[config.channel_name] = config
            _ = self.writer_node(config.channel_name, channels.get_channel(config.channel_name))

    def _enrich(self, val: Any):
        if isinstance(val, list):
            return json.dumps([self._enrich(x) for x in val])
        elif isinstance(val, dict):
            return json.dumps({k: self._enrich(v) for k, v in val.items()})
        elif isinstance(val, datetime):
            return val.strftime("%Y-%m-%d %H:%M:%S.%f")
        elif not isinstance(val, (int, float, str)):  # kind of lazy but non prims get moved to str
            return str(val)

        return val

    def _write_single_item(self, channel_name: str, x: Any):
        from sqlalchemy import create_engine, exc, text

        schema = self.channel_map.get(channel_name)
        xjson = x.to_dict()

        insert_cols = []
        insert_values = []

        fields = schema.fields
        if len(fields) == 0:
            fields = xjson.keys()

        for k, v in schema.augmentation_fields.items():
            insert_cols.append(k)
            insert_values.append(v)

        for field in fields:
            val = xjson.get(field)
            if val is not None:
                insert_cols.append(schema.rename_fields.get(field, field))  # if the field is in the rename_fields rename it otherwise return it
                insert_values.append(self._enrich(val))

        insert_q = f"""INSERT INTO {schema.table} ({", ".join(insert_cols)}) values ({str(insert_values).strip("[]")})"""

        for _ in range(self.n_tries):
            try:
                with self.engine.begin() as connection:
                    connection.execute(text(insert_q))
                logger.info(insert_q)
                return
            except exc.OperationalError as e:
                logger.error(f"Cannot execute query due to {e}, retrying...")
                self.engine = create_engine(self.cnx_details.get_cnx_string(), pool_pre_ping=True)
        if self.fail_after_retry:
            raise exc.OperationalError(f"Unable to insert query {insert_q} after {self.n_tries} tries.")

    @csp.node
    def writer_node(self, channel_name: str, channel: ts["T"]):
        if csp.ticked(channel):
            if isinstance(channel, list):
                for item in channel:
                    self._write_single_item(channel_name, item)
            else:
                self._write_single_item(channel_name, channel)



from datetime import datetime, timedelta, timezone
from typing import List
from unittest import mock

import csp
import datadog
from csp import Enum, ts

from csp_gateway.server.gateway import Gateway, GatewayChannels, GatewayModule
from csp_gateway.server.modules.logging.datadog import PublishDatadog
from csp_gateway.server.modules.logging.util import (
    MonitoringEvent,
    MonitoringMetric,
)
from csp_gateway.testing import GatewayTestHarness


class DatadogTestEnum(Enum):
    GOOD = Enum.auto()
    BAD = Enum.auto()


class DatadogTestChannels(GatewayChannels):
    event: ts[List[MonitoringEvent]] = None
    metric: ts[List[MonitoringMetric]] = None


class DatadogTestModule(GatewayModule):
    interval: timedelta = timedelta(seconds=1)

    @csp.node
    def subscribe_events(
        self,
        trigger: ts[bool],
    ) -> ts[List[MonitoringEvent]]:
        with csp.state():
            last_x = 0
        if csp.ticked(trigger):
            last_x += 1
            return [
                MonitoringEvent(
                    title=f"Event #{last_x}",
                    text=f"Disconnected {last_x}",
                    alert_type=MonitoringEvent.AlertType.error,
                    event_type="disconnect error",
                ),
                MonitoringEvent(
                    title=f"Event #{last_x}",
                    text=f"Cancel order {last_x}",
                    alert_type=MonitoringEvent.AlertType.warning,
                    event_type="cancel order",
                ),
                MonitoringEvent(
                    title=f"Event #{last_x}",
                    text=f"Completed order {last_x}",
                    alert_type=MonitoringEvent.AlertType.success,
                    event_type="completed order",
                ),
                MonitoringEvent(
                    title=f"Event #{last_x}",
                    text=f"Restricted list request {last_x}",
                    alert_type=MonitoringEvent.AlertType.info,
                    event_type="restricted list",
                    tags={"restricted event": f"{last_x}"},
                ),
            ]

    @csp.node
    def subscribe_metrics(
        self,
        trigger: ts[bool],
    ) -> ts[List[MonitoringMetric]]:
        with csp.state():
            last_x = 1.0
        if csp.ticked(trigger):
            last_x += 1
            return [
                MonitoringMetric(
                    metric="cancelled.orders",
                    metric_type=MonitoringMetric.MetricType.count,
                    value=1,
                    tags={"cancelled order": f"{last_x}"},
                ),
                MonitoringMetric(
                    metric="submitted.orders",
                    metric_type=MonitoringMetric.MetricType.gauge,
                    value=10,
                    tags={"submitted order": f"{last_x}"},
                ),
                MonitoringMetric(
                    metric="rejected.orders",
                    metric_type=MonitoringMetric.MetricType.rate,
                    value=100,
                    tags={"rejected order": f"{last_x}"},
                ),
            ]

    def connect(self, channels: DatadogTestChannels):
        # Create some CSP data streams
        events = self.subscribe_events(csp.timer(interval=self.interval, value=True))
        metrics = self.subscribe_metrics(csp.timer(interval=self.interval, value=True))
        # Channels set via `set_channel`
        channels.set_channel(DatadogTestChannels.event, events)
        channels.set_channel(DatadogTestChannels.metric, metrics)


def test_monitoring_init():
    event = MonitoringEvent(
        title="Event #1",
        text="Restricted list request 1",
        alert_type=MonitoringEvent.AlertType.info,
        event_type="restricted list",
        tags={"fav number": 10.27},
    )
    assert event.tags["fav number"] == "10.27"

    metric = MonitoringMetric(
        metric="rejected.orders",
        metric_type=MonitoringMetric.MetricType.rate,
        value=100,
        tags={"enum": DatadogTestEnum.GOOD},
    )
    assert isinstance(metric.tags["enum"], str)


def test_datadog_bad_connection():
    # no connection provided should raise API error
    h = GatewayTestHarness(test_channels=["event", "metric"])
    gateway = Gateway(
        modules=[
            h,
            DatadogTestModule(),
            PublishDatadog(
                dd_tags={"user": "test0123", "trading_area": "ABCD"},
                events_channel="event",
                metrics_channel="metric",
            ),
        ],
        channels=DatadogTestChannels(),
    )
    h.delay(timedelta(seconds=1))
    csp.run(
        gateway.graph,
        starttime=datetime.now(timezone.utc),
        endtime=timedelta(seconds=5),
    )


@mock.patch("datadog.initialize")
@mock.patch.object(datadog.api.Event, "create")
@mock.patch.object(datadog.api.Metric, "send")
def test_datadog(mock_send_metric, mock_event_create, mock_initialize):
    mock_event_create.return_value = {"status": "ok"}

    h = GatewayTestHarness(test_channels=["event", "metric"])
    gateway = Gateway(
        modules=[
            h,
            DatadogTestModule(),
            PublishDatadog(
                dd_tags={"user": "test0123", "trading_area": "ABCD"},
                events_channel="event",
                metrics_channel="metric",
            ),
        ],
        channels=DatadogTestChannels(),
    )
    h.delay(timedelta(seconds=1))

    def assert_func(event_ticks):
        assert len(event_ticks) > 0
        for _, events in event_ticks:
            for e in events:
                # We are asserting here that to_datadog() has been called on the event. to_datadog() mutates the
                # MonitoringBase. There are tests in downstream libraries (e.g. cubist-trading) that expects to_datadog()
                # to be called in the main graph thread. If this assertion fails, the unit tests in downstream libraries
                # might be broken.
                assert "monitoring_id" in e.tags

    h.assert_ticked_values("event", assert_func)
    csp.run(
        gateway.graph,
        starttime=datetime.now(timezone.utc),
        endtime=timedelta(seconds=5),
    )

    event_keys = (
        "title",
        "text",
        "alert_type",
        "aggregation_key",
        "date_happened",
        "source_type_name",
        "user",
        "trading_area",
        "monitoring_id",
    )

    init_event = mock_event_create.call_args_list[0]
    init_dict = init_event.kwargs
    tags = dict(x.split(":") for x in init_dict.pop("tags"))
    init_dict.update(tags)
    for key in event_keys:
        assert key in init_dict
    del init_dict["date_happened"]
    del init_dict["monitoring_id"]
    assert init_dict == {
        "title": "PublishDatadog started",
        "text": "PublishDatadog started",
        "alert_type": "success",
        "aggregation_key": "connection init",
        "source_type_name": "python",
        "user": "test0123",
        "trading_area": "ABCD",
    }

    test_events = mock_event_create.call_args_list[1:]

    # assert event submissions
    # 15 ticks * 4 types of events = 20
    assert len(test_events) == 20
    event_map = {}
    for res in test_events:
        res_dict = res.kwargs
        tags = dict(x.split(":") for x in res_dict.pop("tags"))
        res_dict.update(tags)

        # assert all keys are in the list
        assert [key in res_dict for key in event_keys] == [True] * 9
        if res_dict["aggregation_key"] == "restricted list":
            assert "restricted event" in res_dict

        # count tag values
        for v in res_dict.values():
            if v not in event_map:
                event_map[v] = 0
            event_map[v] += 1

    for i in range(1, 6):
        assert event_map[f"Event #{i}"] == 4

    for v in MonitoringEvent.AlertType:
        assert event_map[v.value] == 5

    for v in ("python", "test0123", "ABCD"):
        assert event_map[v] == 20

    for v in ("disconnect error", "restricted list", "cancel order", "completed order"):
        assert event_map[v] == 5

    # assert metric submissions
    # 5 ticks * 3 types of metrics = 15
    assert mock_send_metric.call_count == 15
    metric_keys = (
        "metric",
        "type",
        "points",
        "user",
        "trading_area",
        "monitoring_id",
    )
    metric_map = {}
    sum = 0
    for res in mock_send_metric.call_args_list:
        res_dict = res.kwargs
        tags = dict(x.split(":") for x in res_dict.pop("tags"))
        res_dict.update(tags)

        # assert all keys are in the list
        assert [key in res_dict for key in metric_keys] == [True] * 6

        # count tag values
        for v in res_dict.values():
            if isinstance(v, List):
                sum += int(v[0][1])
                continue
            else:
                if v not in metric_map:
                    metric_map[v] = 0
                metric_map[v] += 1

    # assert values
    assert sum == 555

    for v in (
        "cancelled.orders",
        "submitted.orders",
        "rejected.orders",
        "count",
        "gauge",
        "rate",
    ):
        assert metric_map[v] == 5

    for v in ("test0123", "ABCD"):
        assert metric_map[v] == 15



from datetime import datetime, timedelta

import csp
import pytest

from csp_gateway.server.modules import AddChannelsToGraphOutput
from csp_gateway.testing import GatewayTestHarness
from csp_gateway.testing.shared_helpful_classes import MyGateway, MyGatewayChannels, MyStruct


@pytest.mark.parametrize("make_invalid", (True, False))
def test_delay(make_invalid):
    channels = [
        MyGatewayChannels.my_channel,
    ]
    h = GatewayTestHarness(test_channels=channels)

    # Engine start
    h.send(MyGatewayChannels.my_channel, MyStruct())
    h.assert_ticked(MyGatewayChannels.my_channel, 1)

    # timedelta delay
    h.advance(delay=timedelta(seconds=1))
    h.send(MyGatewayChannels.my_channel, MyStruct())

    # timedelta delay on timedelta delay
    h.advance(delay=timedelta(seconds=2))
    h.send(MyGatewayChannels.my_channel, MyStruct())

    # datetime delay
    h.advance(delay=datetime(2020, 1, 2))
    h.send(MyGatewayChannels.my_channel, MyStruct())

    # timedelta delay on datetime delay
    h.advance(delay=timedelta(seconds=5))
    h.send(MyGatewayChannels.my_channel, MyStruct())

    if make_invalid:
        # Jumping back in time.
        h.advance(delay=datetime(2020, 1, 2, 0, 0, 1))

    gateway = MyGateway(
        modules=[h, AddChannelsToGraphOutput()],
        channels=MyGatewayChannels(),
    )

    if make_invalid:
        with pytest.raises(ValueError):
            csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(3))
    else:
        res = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(3))

        assert "my_channel" in res
        assert len(res["my_channel"]) == 5
        expected_times = [
            datetime(2020, 1, 1),
            datetime(2020, 1, 1, 0, 0, 1),
            datetime(2020, 1, 1, 0, 0, 3),
            datetime(2020, 1, 2),
            datetime(2020, 1, 2, 0, 0, 5),
        ]
        for (actual_time, _), expected_time in zip(res["my_channel"], expected_times):
            assert actual_time == expected_time


def test_delay_jump_straight_away():
    channels = [
        MyGatewayChannels.my_channel,
    ]
    h = GatewayTestHarness(test_channels=channels)

    # datetime delay
    h.advance(delay=datetime(2020, 1, 2))
    h.send(MyGatewayChannels.my_channel, MyStruct())

    # timedelta delay on datetime delay
    h.advance(delay=timedelta(seconds=5))
    h.send(MyGatewayChannels.my_channel, MyStruct())

    gateway = MyGateway(
        modules=[h, AddChannelsToGraphOutput()],
        channels=MyGatewayChannels(),
    )
    res = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(3))

    assert "my_channel" in res
    assert len(res["my_channel"]) == 2
    expected_times = [
        datetime(2020, 1, 2),
        datetime(2020, 1, 2, 0, 0, 5),
    ]
    for (actual_time, _), expected_time in zip(res["my_channel"], expected_times):
        assert actual_time == expected_time



import os
import os.path
from typing import Optional

from fastapi import HTTPException, Request
from fastapi.responses import HTMLResponse, StreamingResponse
from hydra.core.hydra_config import HydraConfig

try:
    # conditional on libmagic being installed on the machine
    from magic import Magic
except ImportError:
    Magic = None

from csp_gateway.server import GatewayChannels, GatewayModule

# separate to avoid circular
from csp_gateway.server.web import GatewayWebApp


class MountOutputsFolder(GatewayModule):
    dir: Optional[str] = None

    def connect(self, channels: GatewayChannels) -> None:
        if self.dir is None:
            if HydraConfig.initialized():
                self.dir = os.path.abspath(os.path.join(HydraConfig.get().runtime.output_dir, "..", "..", ".."))
            else:
                self.dir = os.path.abspath(os.path.join(os.getcwd(), "outputs"))

    def rest(self, app: GatewayWebApp) -> None:
        app_router = app.get_router("app")
        mime = Magic(mime=True) if Magic else None

        # TODO subselect
        @app_router.get("/outputs/{full_path:path}", response_class=HTMLResponse, tags=["Utility"])
        def browse_logs(full_path: str, request: Request):
            """
            This endpoint is a small webpage for browsing the [hydra](https://github.com/facebookresearch/hydra)
            output logs and configuration settings of the running application.
            """
            file_or_dir = self.dir
            if full_path:
                file_or_dir = os.path.join(file_or_dir, full_path)
            if os.path.abspath(file_or_dir).startswith(self.dir) and os.path.exists(file_or_dir):
                if os.path.isdir(file_or_dir):
                    files = os.listdir(file_or_dir)
                    files_paths = sorted([f"{request.url._url}/{f}".replace("outputs//", "outputs/") for f in files])
                    return app.templates.TemplateResponse(
                        "files.html.j2", {"request": request, "files": files_paths, "pid": os.getpid()}, media_type="text/html"
                    )

                def iterfile():  #
                    with open(file_or_dir, "rb") as fp:
                        yield from fp

                if file_or_dir.endswith((".log", ".txt")):
                    # NOTE: so viewable in browser, magic is guessing wrong type
                    media_type = "text/plain; charset=utf-8"
                elif mime:
                    media_type = mime.from_file(file_or_dir)
                else:
                    media_type = None

                return StreamingResponse(iterfile(), media_type=media_type)
            raise HTTPException(status_code=404, detail="Not found: {}".format(request.url._url))



from datetime import datetime, timedelta
from typing import Dict, List

import csp
import numpy as np
import pytest
from csp import ts
from csp.typing import Numpy1DArray

from csp_gateway import PrintChannels
from csp_gateway.testing.shared_helpful_classes import (
    GatewayChannels,
    GatewayModule,
    MyEnum,
    MyGateway,
    State,
)


class MyPrintGatewayChannels(GatewayChannels):
    my_static: float = 0.0
    my_static_dict: Dict[str, float] = {}
    my_static_list: List[str] = []
    my_channel: ts[int] = None
    s_my_channel: ts[State[int]] = None
    my_list_channel: ts[[int]] = None
    s_my_list_channel: ts[State[int]] = None
    my_enum_basket: Dict[MyEnum, ts[int]] = None
    my_str_basket: Dict[str, ts[int]] = None
    my_enum_basket_list: Dict[MyEnum, ts[[int]]] = None
    my_array_channel: ts[Numpy1DArray[float]] = None


class MyPrintSetModule(GatewayModule):
    my_data: ts[int]
    my_data2: ts[int]
    my_list_data: List[int]
    by_key: bool = True

    def dynamic_keys(self):
        return {MyPrintGatewayChannels.my_str_basket: ["my_key", "my_key2"]}

    def connect(self, channels: MyPrintGatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        channels.set_channel(MyPrintGatewayChannels.my_channel, self.my_data)
        channels.set_channel(
            MyPrintGatewayChannels.my_list_channel,
            csp.const(self.my_list_data),
        )
        channels.add_send_channel(MyPrintGatewayChannels.my_channel)
        channels.set_state(MyPrintGatewayChannels.my_channel, "id")
        channels.add_send_channel(MyPrintGatewayChannels.my_list_channel)
        channels.set_state(MyPrintGatewayChannels.my_list_channel, "id")

        channels.set_channel(MyPrintGatewayChannels.my_array_channel, csp.const(np.array([1.0, 2.0])))

        if self.by_key:
            channels.set_channel(MyPrintGatewayChannels.my_enum_basket, self.my_data, MyEnum.TWO)
            channels.set_channel(MyPrintGatewayChannels.my_str_basket, self.my_data2, "my_key2")
            channels.set_channel(
                MyPrintGatewayChannels.my_enum_basket_list,
                csp.const(self.my_list_data),
                MyEnum.ONE,
            )
        else:
            channels.set_channel(
                MyPrintGatewayChannels.my_enum_basket,
                {
                    MyEnum.TWO: self.my_data,
                },
            )
            channels.set_channel(
                MyPrintGatewayChannels.my_str_basket,
                {
                    "my_key2": self.my_data2,
                },
            )
            channels.set_channel(
                MyPrintGatewayChannels.my_enum_basket_list,
                {
                    MyEnum.ONE: csp.const(self.my_list_data),
                },
            )


@pytest.mark.parametrize("by_key", [True, False])
def test_PrintChannels(by_key, capsys):
    setter = MyPrintSetModule(
        my_data=csp.const(9),
        my_data2=csp.const(8),
        my_list_data=[8, 9],
        by_key=by_key,
    )
    getter = PrintChannels()
    gateway = MyGateway(modules=[getter, setter], channels=MyPrintGatewayChannels())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    captured = capsys.readouterr()
    printed_set = set(captured.out.split("\n"))
    expected_results = [
        "2020-01-01 00:00:00 my_array_channel:[1. 2.]",
        "2020-01-01 00:00:00 my_str_basket:{'my_key2': 8}",
        "2020-01-01 00:00:00 my_enum_basket:{<MyEnum.TWO: 2>: 9}",
        "2020-01-01 00:00:00 my_enum_basket_list:{<MyEnum.ONE: 1>: [8, 9]}",
        "2020-01-01 00:00:00 my_channel:9",
        "2020-01-01 00:00:00 my_list_channel:[8, 9]",
        "",
    ]
    assert len(expected_results) == len(printed_set)
    for result in expected_results:
        assert result in printed_set


@pytest.mark.parametrize("by_key", [True, False])
def test_specific_channels_PrintChannels(by_key, capsys):
    setter = MyPrintSetModule(
        my_data=csp.const(9),
        my_data2=csp.const(8),
        my_list_data=[8, 9],
        by_key=by_key,
    )
    getter = PrintChannels(
        selection=[
            MyPrintGatewayChannels.my_array_channel,
            MyPrintGatewayChannels.my_enum_basket,
            MyPrintGatewayChannels.my_static_list,
        ]
    )
    gateway = MyGateway(modules=[getter, setter], channels=MyPrintGatewayChannels())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    captured = capsys.readouterr()
    printed_set = set(captured.out.split("\n"))
    expected_results = [
        "2020-01-01 00:00:00 my_array_channel:[1. 2.]",
        "2020-01-01 00:00:00 my_enum_basket:{<MyEnum.TWO: 2>: 9}",
        "",
    ]
    assert len(expected_results) == len(printed_set)
    for result in expected_results:
        assert result in printed_set



from datetime import datetime, timedelta
from typing import List, TypeVar, Union, get_args, get_origin  # noqa: TYP001

import csp
from csp import Outputs, ts
from csp.impl.types.tstype import TsType, isTsType
from deprecation import deprecated

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

__all__ = (
    "_get_dict_basket_key_type",
    "_get_dict_basket_value_tstype",
    "_get_dict_basket_value_type",
    "_is_dict_basket",
    "_is_list_basket",
    "get_dict_basket_key_type",
    "get_dict_basket_value_tstype",
    "get_dict_basket_value_type",
    "is_dict_basket",
    "is_list_basket",
    "get_args",
    "get_origin",
    "set_alarm_and_fetch_alarm_time",
    "to_list",
)


def is_dict_basket(val: type) -> bool:
    return get_origin(val) is dict and isTsType(get_args(val)[1])


def is_list_basket(val: type) -> bool:
    return get_origin(val) is list and isTsType(get_args(val)[0])


def get_dict_basket_key_type(val: type) -> type:
    if not is_dict_basket(val):
        raise TypeError(f"object is not a Dict Basket but is of type: {type(val)}")
    return get_args(val)[0]


def get_dict_basket_value_tstype(val: type) -> TsType:
    if not is_dict_basket(val):
        raise TypeError(f"object is not a Dict Basket but is of type: {type(val)}")
    return get_args(val)[1]


def get_dict_basket_value_type(val: type) -> type:
    return get_dict_basket_value_tstype(val).typ


@deprecated(details="Use is_dict_basket instead.")
def _is_dict_basket(val: type) -> bool:
    return is_dict_basket(val)


@deprecated(details="Use is_list_basket instead.")
def _is_list_basket(val: type) -> bool:
    return is_list_basket(val)


@deprecated(details="Use get_dict_basket_key_type instead.")
def _get_dict_basket_key_type(val: type) -> type:
    return get_dict_basket_key_type(val)


@deprecated(details="Use get_dict_basket_value_tstype instead.")
def _get_dict_basket_value_tstype(val: type) -> TsType:
    return get_dict_basket_value_tstype(val)


@deprecated(details="Use get_dict_basket_value_type instead.")
def _get_dict_basket_value_type(val: type) -> type:
    return get_dict_basket_value_type(val)


@csp.node
def to_list(x: ts["T"]) -> ts[List["T"]]:
    if csp.ticked(x):
        return [x]


@csp.node
def set_alarm_and_fetch_alarm_time(time: Union[datetime, timedelta]) -> Outputs(alarm_time=ts[datetime], alarm_ticked=ts[bool]):
    with csp.alarms():
        engine_start: ts[bool] = csp.alarm(bool)
        alarm: ts[bool] = csp.alarm(bool)
    with csp.state():
        s_time = None

    with csp.start():
        csp.schedule_alarm(engine_start, timedelta(), True)
        csp.schedule_alarm(alarm, time, True)
        if isinstance(time, datetime):
            s_time = time
        else:
            s_time = csp.now() + time

    if csp.ticked(engine_start):
        csp.output(alarm_time=s_time)
    if csp.ticked(alarm):
        csp.output(alarm_ticked=True)



from datetime import datetime, timedelta
from typing import Dict, List

import csp
import pytest

from csp_gateway.utils import (
    get_dict_basket_key_type,
    get_dict_basket_value_tstype,
    get_dict_basket_value_type,
    is_dict_basket,
    is_list_basket,
    set_alarm_and_fetch_alarm_time,
    to_list,
)


def test_dict_basket_type_checks():
    test_dict_typ = Dict[int, str]
    test_dict_basket_typ = Dict[int, csp.ts[int]]
    assert not is_dict_basket(test_dict_typ)
    assert is_dict_basket(test_dict_basket_typ)

    with pytest.raises(TypeError):
        get_dict_basket_key_type(test_dict_typ)

    with pytest.raises(TypeError):
        get_dict_basket_value_tstype(test_dict_typ)

    assert get_dict_basket_key_type(test_dict_basket_typ) is int
    assert get_dict_basket_value_tstype(test_dict_basket_typ) is csp.ts[int]
    assert get_dict_basket_value_type(test_dict_basket_typ) is int


def test_is_list_basket():
    test_list_typ = List[int]
    test_list_basket_typ = List[csp.ts[int]]
    assert not is_list_basket(test_list_typ)
    assert is_list_basket(test_list_basket_typ)


def test_to_list():
    out = csp.run(
        to_list,
        csp.const(9),
        starttime=datetime(2020, 1, 1),
        endtime=timedelta(1),
    )
    assert out[0] == [(datetime(2020, 1, 1), [9])]


def test_set_alarm_and_fetch_alarm_time():
    out = csp.run(
        set_alarm_and_fetch_alarm_time,
        timedelta(),
        starttime=datetime(2020, 1, 1),
        endtime=timedelta(1),
    )
    assert out["alarm_time"] == [(datetime(2020, 1, 1), datetime(2020, 1, 1))]
    assert out["alarm_ticked"] == [(datetime(2020, 1, 1), True)]

    out = csp.run(
        set_alarm_and_fetch_alarm_time,
        timedelta(minutes=1),
        starttime=datetime(2020, 1, 1),
        endtime=timedelta(1),
    )
    assert out["alarm_time"] == [(datetime(2020, 1, 1), datetime(2020, 1, 1, 0, 1))]
    assert out["alarm_ticked"] == [(datetime(2020, 1, 1, 0, 1), True)]

    out = csp.run(
        set_alarm_and_fetch_alarm_time,
        datetime(2020, 1, 7),
        starttime=datetime(2020, 1, 1),
        endtime=timedelta(days=14),
    )
    assert out["alarm_time"] == [(datetime(2020, 1, 1), datetime(2020, 1, 7))]
    assert out["alarm_ticked"] == [(datetime(2020, 1, 7), True)]

    out = csp.run(
        set_alarm_and_fetch_alarm_time,
        datetime(2021, 1, 1),
        starttime=datetime(2020, 1, 1),
        endtime=timedelta(days=14),
    )
    assert out["alarm_time"] == [(datetime(2020, 1, 1), datetime(2021, 1, 1))]
    assert out["alarm_ticked"] == []



import abc
import datetime
import logging
import threading
import typing
from collections import deque
from enum import Enum as PyEnum
from functools import lru_cache
from pprint import pformat
from typing import Any, Deque, Dict, List, Tuple, Union

import csp
import duckdb
import fsspec
import numpy
import orjson
from atomic_counter import Counter
from ccflow.enums import BaseEnum as CoreBaseEnum, Enum as CoreEnum
from csp import Struct, ts
from csp.impl.enum import EnumMeta as CspEnumMeta
from pydantic import BaseModel
from typing_extensions import override

if typing.TYPE_CHECKING:
    from csp_gateway.utils import Query

_DUCKDB_BUFFER_THRESHOLD_ORIGINAL = 1000000
_DUCKDB_BUFFER_THRESHOLD_CURRENT = _DUCKDB_BUFFER_THRESHOLD_ORIGINAL
_USE_DUCKDB_STATE = True

log = logging.getLogger(__name__)

__all__ = [
    "StateType",
    "State",
    "DefaultState",
    "DuckDBState",
    "build_track_state_node",
    "modify_buffer_threshold",
    "restore_buffer_threshold",
    "enable_duckdb_state",
    "disable_duckdb_state",
]


# NOTE: Consider using a singleton class for configuration settings for DuckDB
def modify_buffer_threshold(new_threshold: int) -> None:
    """Set DuckDB buffer threshold to a new value"""

    global _DUCKDB_BUFFER_THRESHOLD_CURRENT
    if new_threshold < 0:
        raise ValueError("Threshold should be a non-negative number")
    _DUCKDB_BUFFER_THRESHOLD_CURRENT = new_threshold


def restore_buffer_threshold() -> None:
    """Reset DuckDB buffer threshold to original value"""

    global _DUCKDB_BUFFER_THRESHOLD_CURRENT
    _DUCKDB_BUFFER_THRESHOLD_CURRENT = _DUCKDB_BUFFER_THRESHOLD_ORIGINAL


def enable_duckdb_state() -> None:
    """Enable using DuckDBState as State class specialization"""

    global _USE_DUCKDB_STATE
    _USE_DUCKDB_STATE = True


def disable_duckdb_state() -> None:
    """Disable using DuckDBState as State class specialization"""

    global _USE_DUCKDB_STATE
    _USE_DUCKDB_STATE = False


class StateType(CoreEnum):
    UNKNOWN = 0
    DEFAULT = 1
    DUCKDB = 2


class StateQueryParams(BaseModel):
    # TODO
    ...


class BaseState(abc.ABC):
    """Abstract class for State"""

    @abc.abstractmethod
    def query(self, query: "Query" = None) -> List[Any]: ...

    @abc.abstractmethod
    def __iter__(self) -> Any: ...

    @abc.abstractmethod
    def insert(self, record: Any) -> None: ...

    @abc.abstractmethod
    def __repr__(self) -> str: ...


class DefaultState(BaseState):
    def __init__(self, keyby: Union[Tuple[str, ...], str]) -> None:
        super().__init__()
        # TODO psp?
        self._records: Dict = {}
        self._keyby = keyby if isinstance(keyby, tuple) else (keyby,)

    @override
    def query(self, query: "Query" = None) -> List[Any]:
        # TODO
        max_depth = len(self._keyby)
        flattened = []

        to_visit: Deque[Tuple[list, dict, int]] = deque()
        to_visit.append(([], self._records, 0))

        while to_visit:
            key, structure, depth = to_visit.pop()

            if depth == max_depth:
                flattened.append(key + [structure])
            else:
                for subkey in structure:
                    to_visit.append((key + [subkey], structure[subkey], depth + 1))

        # Since elements of the key might be None, need to be careful.
        # We choose to have None records appear at the beginning
        # Taking str(subkey) handles the Enum case, as < not implemented for Enums
        flattened.sort(key=lambda e: [(subkey is not None, str(subkey)) for subkey in e[:-1]])
        to_return = [_[-1] for _ in flattened]

        # TODO parse query
        if query and query.filters:
            to_return = query.calculate(to_return)
        return to_return

    @override
    def __iter__(self) -> Any:
        for _ in self._records:
            yield _

    @override
    def insert(self, record: Any) -> None:
        place = self._records

        for subkey in self._keyby:
            # extract the key from the record
            subkey_to_use = getattr(record, subkey, None)

            if subkey == self._keyby[-1]:
                # Put the element there if last
                place[subkey_to_use] = record

            if subkey_to_use not in place:
                place[subkey_to_use] = {}

            place = place[subkey_to_use]

    @override
    def __repr__(self) -> str:
        return pformat(self.query())


class DuckDBState(object):
    """State implementation using DuckDB as the object store and query engine"""

    # NOTE: The duckdb system here is pretty fragile, we should think of building a generict DuckDB interface
    # so that other parts of the codebase and use duckdb through the generic interface

    # NOTE: DuckDB is an OLAP database and not meant to be used as a OLTP database for frequent inserts.
    # We use DuckDB here for its strong analytics capabilities. Row by row insertions into DuckDB are not
    # recommended and will cause major slow downs, instead we bulk load new records into duckdb at certain
    # events: on query call, <more to be added>. This ensures that insert path is fast.
    # For bulk loading, we push json versions of the records into a memory base file. When a bulk loading
    # event occurs, we load the data from the memory file and truncate it.

    # Counter for creating new tables ids for the same type object. This is needed to create an independent
    # table for each instance of the DuckDBState class for any particular type
    TABLE_ID = Counter(1)

    def __init__(self, typ: Struct, keyby: Union[Tuple[str, ...], str], schema: Dict[str, str] = {}) -> None:
        super().__init__()
        # Type of record
        self._typ = typ
        # Table name for this instance of DuckDBState[<self._typ>] class
        self._table_name = f"{typ.__name__}_{DuckDBState.TABLE_ID.next()}"
        # Name of the ID column
        self._id_name = "duck_id"
        # Name of the record column
        self._col_name = "state"
        # Store the schema
        self._schema = schema
        self._schema_str = orjson.dumps(self._schema).decode()
        # NOTE: We make a new connection object for each state object. Using the same state connection across
        #  multiple state objects leads to issues in DuckDB's python API related to pending query results not
        #  being fully executed. This should be okay since we are doing this once during the initialization
        self._con = duckdb.connect()
        self._keyby = keyby
        # ID generator for new records. We cannot rely on the id in the record as that might or might not exist
        self._obj_id_generator = Counter(1)
        # Memory filename to use for bulk loading data into DuckDB
        self._mem_file_name = f"{self._table_name}.json"
        # Tree like structure to store the ids for records using the keys in self._keyby
        self._key_to_id = {}
        # Temporary structure to buffer records before bulk loading into DuckDB
        self._id_record_buffer = {}
        # Lock to prevent race conditions as multiple threads modify the record buffer
        self._buffer_lock = threading.Lock()
        # Lock to prevent race conditions as multiple threads modify the DB
        self._db_lock = threading.Lock()
        self._query_lock = threading.Lock()
        # Object store to recover the original struct. This links our generated ID with the record python object
        # Needed because there is no way to cleanly build the python object from the json dictionary
        self._obj_store = {}

        # Pre-build the group by part of the query, since the keyby values are passed during initialization
        # This is based on the keyby values passed during initialization of the state instance
        key_cols = []
        if isinstance(self._keyby, str):
            self._keyby = (self._keyby,)
        if isinstance(self._keyby, tuple):
            for keyby_term in self._keyby:
                if not DuckDBState.check_attr_schema(keyby_term, self._schema):
                    log.warning(f"Ignoring keyby with {keyby_term}, column not found")
                    continue
                key_cols.append(f"{self._col_name}.{keyby_term}")
        if key_cols:
            key_cols_str = ", ".join(key_cols)
            sort_cols_str = ", ".join([f"{item} ASC NULLS FIRST" for item in key_cols])
            self._group_str = f"GROUP BY {key_cols_str} ORDER BY {sort_cols_str}"
        else:
            self._group_str = ""

        # Create a type for the state object
        duckdb_type = self._con.sql(f"SELECT typeof(json_transform('{{}}', '{self._schema_str}'))").fetchall()[0][0]

        # Create the table
        self._con.sql(
            f"CREATE OR REPLACE TABLE '{self._table_name}' \
                ({self._id_name} BIGINT PRIMARY KEY, {self._col_name} {duckdb_type})"
        )

        # Construct column structure for json parsing
        self._columns_duckdb = str(f"{{'{self._id_name}':'BIGINT','{self._col_name}':'{duckdb_type}'}}")

        # Open the memory file for buffering data before bulk loading
        self._mem_file = fsspec.filesystem("memory").open(self._mem_file_name, "wb")

        # Register memory filesystem so that duckdb can read from memory
        self._con.register_filesystem(fsspec.filesystem("memory"))

    # TODO: Improve this to support querying dicts and lists as well, currently it only
    # works with structs and primitive types
    @classmethod
    def check_attr_schema(cls, attr: str, schema: dict) -> bool:
        """Check if the attr being queried exists in the schema"""

        keys = attr.split(".")
        cur = schema
        try:
            for key in keys:
                cur = cur[key]
        except (TypeError, KeyError):
            return None
        if isinstance(cur, dict):
            return None
        else:
            return cur

    def query_schema(self) -> str:
        """Returns the schema for the duckdb table"""
        return self._schema_str

    #  TODO: Make Query hashable so that we can add @lru_cache() here, to avoid reconstructing an already seen query
    def construct_query(self, query: "Query") -> str:
        """Construct the query from the passed query object, and validate the query attributes by comparing to schema"""

        # Construct the filter part of the query
        conds = []
        # TODO: This is just a copy of the Filter behaviour, find a cleaner way so as to NOT duplicate code
        #  and provide a more generic query functionality
        if query:
            for filter in query.filters:
                attr_type = DuckDBState.check_attr_schema(filter.attr, self._schema)
                if not attr_type:
                    log.warning(f"Ignoring filter with {filter.attr}, column not found")
                    continue
                cond = "TRUE"
                if filter.by.value is not None:
                    by_val = filter.by.value
                    if isinstance(by_val, str):
                        by_val = f"'{by_val}'"
                    cond = f"{self._col_name}.{filter.attr} {filter.by.where} CAST({by_val} AS {attr_type})"
                elif filter.by.when is not None:
                    cond = f"{self._col_name}.{filter.attr} {filter.by.where} CAST('{filter.by.when}' AS TIMESTAMP)"
                elif filter.by.attr:
                    if not DuckDBState.check_attr_schema(filter.by.attr, self._schema):
                        log.warning(f"Ignoring filter with {filter.by.attr}, column not found")
                        continue
                    cond = f"{self._col_name}.{filter.attr} {filter.by.where} {self._col_name}.{filter.by.attr}"
                conds.append(cond)
        if conds:
            filter_str = "WHERE " + " AND ".join(conds)
        else:
            filter_str = ""

        # Combine the filter and groupby to get the final query
        if self._group_str:
            duckdb_query = f"SELECT last({self._id_name}) FROM '{self._table_name}' {filter_str} {self._group_str}"
        else:
            duckdb_query = f"SELECT {self._id_name} FROM '{self._table_name}' {filter_str}"
        return duckdb_query

    def load_data(self) -> None:
        with self._db_lock:
            with self._buffer_lock:
                buffer = self._id_record_buffer
                # Clear the record buffer
                self._id_record_buffer = {}

            # Early breakout if nothing new was added
            if len(buffer) == 0:
                return

            # TODO: Move this to a common place
            def default(obj: Any) -> Any:
                if isinstance(obj, set):
                    return list(obj)
                elif isinstance(obj, numpy.ndarray):
                    return obj.tolist()
                elif isinstance(obj, (PyEnum, CoreBaseEnum, CoreEnum)):
                    return obj.name
                else:
                    logging.warning(f"Type({type(obj)}) cannot be json serialized please provide serializer: {obj} serializing to '' for now")
                    return ""

            for id, record in buffer.items():
                # Insert into object store
                self._obj_store[id] = record

                # Convert the record to json
                record_json_str = record.to_json(default)
                record_json = orjson.loads(record_json_str)
                json_dict = {self._id_name: id, self._col_name: record_json}
                json_str = orjson.dumps(json_dict)

                # Write to buffer instead of DuckDB because single row inserts are slow
                # by buffering we can bulk load the data making it much faster
                self._mem_file.write(json_str)

            # Flush memory buffer
            self._mem_file.flush()

            # Upsert data into the table
            self._con.sql(
                f"INSERT OR REPLACE INTO {self._table_name}\
                            SELECT * FROM read_json_auto('memory://{self._mem_file_name}',\
                                                        columns = {self._columns_duckdb})"
            )

            # Clear data from the memory file
            self._mem_file.seek(0)
            self._mem_file.truncate(0)

    @override
    def query(self, query: "Query" = None) -> List[Any]:
        """Run the query on the duckdb table and return results"""
        # Build the query to run
        query_str = self.construct_query(query)
        with self._query_lock:
            # Bulk load data from the memory file
            self.load_data()

            # Run query
            res = self._con.sql(query_str).fetchall()

            # Fetch the python objects from the object store using returned ids
            obj_res = [self._obj_store[row[-1]] for row in res]

        return obj_res

    @override
    def __iter__(self) -> Any:
        buf = self.query()
        for _ in buf:
            yield _

    @override
    def insert(self, record: Struct) -> None:
        """Insert new record into the table

        Create a new entry for the record in the table using the schema for the record type,
        and store the record in the object store. The record and the new table row are linked by a
        unique id in the table.
        """

        #  NOTE: Currently this methods keep approximately 2 copies of the record so that the orignal record
        # can be fully recovered and returned. There is work to improve the memory utilization of the
        # state class

        with self._buffer_lock:
            place = self._key_to_id

            obj_id = None
            for subkey in self._keyby:
                # extract the key from the record
                subkey_to_use = getattr(record, subkey, None)

                if subkey == self._keyby[-1]:
                    if subkey_to_use not in place.keys():
                        place[subkey_to_use] = self._obj_id_generator.next()
                    obj_id = place[subkey_to_use]
                    self._id_record_buffer[obj_id] = record

                if subkey_to_use not in place:
                    place[subkey_to_use] = {}

                place = place[subkey_to_use]

        # If the buffer is large enough load into DuckDB
        if len(self._id_record_buffer) >= _DUCKDB_BUFFER_THRESHOLD_CURRENT:
            self.load_data()

    @override
    def __repr__(self) -> str:
        return pformat(self.query())


def get_duckdb_schema_obj(parent: Any, key: Any, cls: Any) -> Tuple[Any, bool]:
    """Create a schema for the passed type object"""

    if not isinstance(cls, type):
        if isinstance(cls, list):
            cls = list
        elif isinstance(cls, dict):
            cls = dict
        elif isinstance(cls, set):
            cls = set
        else:
            return (cls, False)

    if issubclass(cls, Struct):
        # Recursive handling
        return get_duckdb_schema_struct(cls)

    if issubclass(cls, set):
        cls = list

    if issubclass(cls, list):
        #  NOTE: Need to find a clean way to handle different list types (list, numpy arrays, etc)
        #  NOTE: List updates are not supported by duckdb at the moment, so just keep them as strings instead
        #  try:
        #      annotation = parent.__full_metadata_typed__[key]
        #      val_type = typing.get_args(annotation)[0]
        #      new_val_type = get_duckdb_schema_obj(None, None, val_type)
        #      cls = list[val_type]
        #  except (KeyError, IndexError, AttributeError):
        #      log.warning(f"Cannot handle list schema in DuckDB {parent}[{key}]{cls}")
        cls = str
    elif issubclass(cls, dict):
        #  NOTE: Need to find a clean way to handle different dict types (dict, typing.Dict, OrderedDict, etc)
        #  NOTE: Dictionary updates are not supported by duckdb yet, so just keep them as strings instead
        #  try:
        #      annotation = parent.__full_metadata_typed__[key]
        #      key_type = typing.get_args(annotation)[0]
        #      new_key_type = get_duckdb_schema_obj(None, None, key_type)
        #      val_type = typing.get_args(annotation)[1]
        #      new_val_type = get_duckdb_schema_obj(None, None, val_type)
        #      cls = dict[key_type, val_type]
        #  except (KeyError, IndexError):
        #      log.warning(f"Cannot handle dict schema in DuckDB {parent}[{key}]{cls}")
        cls = str
    elif issubclass(cls, (csp.Enum, CoreBaseEnum, CspEnumMeta, PyEnum)):
        # Enums become strings
        cls = str
    elif issubclass(cls, datetime.datetime):
        cls = "datetime"
    elif issubclass(cls, datetime.date):
        cls = "date"
    elif issubclass(cls, datetime.time):
        cls = "time"
    elif issubclass(cls, datetime.timedelta):
        # NOTE: timedelta not supported in duckdb
        cls = str

    try:
        # Convert type to a duckdb type
        return (str(duckdb.typing.DuckDBPyType(cls)), True)
    except Exception:
        # TODO: Be more specific in the exception we need to handle here
        return (cls, False)


def get_duckdb_schema_struct(cls: Struct) -> Tuple[Dict, bool]:
    """Create a partial schema for the struct consisting of parts that can be json_serialized and
    stored in duckdb"""

    orig_type_info = {k: v for k, v in cls.metadata().items() if not k.startswith("_")}
    new_type_info = {}
    use_duckdb = False

    for k, v in orig_type_info.items():
        item_schema, use_duckdb_item = get_duckdb_schema_obj(cls, k, v)
        if use_duckdb_item:
            new_type_info[k] = item_schema
            use_duckdb = True
        else:
            log.warning(f"Cannot support {cls.__name__}.{k}:{v} in duckdb")

    return (new_type_info, use_duckdb)


# NOTE: NEVER access State object directly, always access through the __class_getitem__ API
class State(BaseState):
    def __init__(self, keyby: Union[Tuple[str, ...], str] = ("id",)) -> None:
        """Switch case between different state specializations based on the type of the records"""

        global _USE_DUCKDB_STATE
        try:
            typ = self._typ
            if _USE_DUCKDB_STATE and isinstance(typ, type) and issubclass(typ, Struct):
                schema, use_duckdb = get_duckdb_schema_struct(typ)
                if use_duckdb:
                    self._state_impl = DuckDBState(typ, keyby, schema)
                    self._state_type = StateType.DUCKDB
                    return
        except AttributeError:
            log.warning("Do not create object directly from State use the State[<typ>] API instead for performance reasons")
            log.warning("Using DefaultStateClass")
        self._state_impl = DefaultState(keyby)
        self._state_type = StateType.DEFAULT

    def state_type(self) -> StateType:
        """Return the state type being used internally"""
        return self._state_type

    @override
    def query(self, query: "Query" = None) -> List[Any]:
        # NOTE: These try catch blocks for the experimental duckdb state specialization
        #       Remove them when duckdb state has become stable
        try:
            return self._state_impl.query(query)
        except Exception as err:
            log.error(f"Querying state raised exception {err}")
            return []

    @override
    def __iter__(self) -> Any:
        self._state_impl.__iter__()

    @override
    def insert(self, record: Any) -> None:
        # NOTE: These try catch blocks for the experimental duckdb state specialization
        #       Remove them when duckdb state has become stable
        try:
            self._state_impl.insert(record)
        except Exception as err:
            log.error(f"Inserting state raised exception {err}")

    @override
    def __repr__(self) -> str:
        return self._state_impl.__repr__()

    @classmethod
    @lru_cache()
    def __class_getitem__(cls: type, typ: type) -> Any:
        new_cls = type("State[{}]".format(typ.__name__), (State,), {})
        new_cls._typ = typ
        return new_cls


def build_track_state_node(edge: Any, keyby: Union[str, Tuple[str, ...]]) -> Any:
    @csp.node
    def _track_state_node(  # type: ignore[no-untyped-def]
        ts: ts[edge.tstype.typ], keyby: object
    ) -> ts[State[edge.tstype.typ]]:
        with csp.state():
            s_tracker = State[edge.tstype.typ](keyby)
        if csp.ticked(ts):
            s_tracker.insert(ts)
            csp.output(s_tracker)

    return _track_state_node(edge, keyby)



import socket

import pytest


@pytest.fixture(scope="class")
def free_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        s.listen(1)
        return s.getsockname()[1]



from typing import List, Optional, Set, Union

from ccflow import BaseModel
from csp.impl.types.tstype import isTsType
from pydantic import Field, model_validator

from csp_gateway.server.gateway.csp import Channels, ChannelsType
from csp_gateway.utils import is_dict_basket

__all__ = ("ChannelSelection",)


class ChannelSelection(BaseModel):
    """
    A class to represent channel selection options for filtering channels based on inclusion and exclusion criteria.

    Attributes:
        `include` (Optional[List[str]]): A list of channel names to include in the selection.
            The order here matters
            Defaults to None (include everything).
        `exclude` (Set[str]): A list of channel names to exclude from the selection.
            This overrides anything in `include`
            Defaults to an empty set.

    Methods:
        select_from(channels, static_fields=False, state_channels=False): Returns a list of selected channel names
            based on the inclusion and exclusion criteria, and optional static_fields and
            state_channels flags. The order of channels is based on the order of the channels
            in `include`. State channels always follow their corresponding channels.
            If `include` is None, the order is based on the order of the fields in the channels object.
        validate(v): Validates and coerces the input value to a ChannelSelection instance.
    """

    include: Optional[List[str]] = None
    exclude: Set[str] = Field(default_factory=set)

    @model_validator(mode="before")
    def validate_requires(cls, v):
        if v is None:
            return {}
        if isinstance(v, list):
            return dict(include=list(v))
        return v

    def select_from(
        self,
        channels: Union[Channels, ChannelsType],
        *,
        static_fields: bool = False,  # Select only static fields
        state_channels: bool = False,  # Select only state channels
        all_fields: bool = False,  # Select all fields in include and not in exclude
    ) -> List[str]:
        """
        Select fields from the given channels based on the specified criteria.

        Args:
            channels (Union[Channels, ChannelsType]): The channels to select fields from.
            static_fields (bool, optional): If True, select only static fields. Defaults to False.
            state_channels (bool, optional): If True, select only state channels. Defaults to False.
            all_fields (bool, optional): If True, select all fields in include and not in exclude. Defaults to False.

        Returns:
            List[str]: A list of selected field names.
        """
        names = {}

        if all_fields:
            fields = channels.model_fields if self.include is None else self.include
            return list(dict.fromkeys([field for field in fields if field not in self.exclude]))

        for idx, field in enumerate(channels.model_fields):
            # avoid duplicates
            if field in names:
                continue

            # TODO not this `s_` business...
            # Return state channels or regular channels
            if state_field := field.startswith("s_"):
                if not state_channels:
                    continue
                field = field[2:]
            else:
                if state_channels:
                    continue

            # Check whether static
            outer_type = channels.get_outer_type(field)
            if is_dict_basket(outer_type) or isTsType(outer_type):
                if static_fields:
                    continue
            else:
                if not static_fields:
                    continue

            # Check whether included
            if self.include is not None:
                try:
                    idx = self.include.index(field)
                except ValueError:
                    continue

            # Check whether excluded
            if field in self.exclude:
                continue

            if state_field:
                field = f"s_{field}"

            names[field] = idx

        return [k for k, _ in sorted(names.items(), key=lambda x: x[1])]



import logging
from abc import abstractmethod
from datetime import timezone
from enum import Enum
from typing import Dict, List, Tuple, Union

import orjson
from pydantic import BaseModel
from typing_extensions import override

from csp_gateway.utils.struct.base import GatewayStruct

log = logging.getLogger(__name__)

__all__ = (
    "DatadogLevel",
    "MonitoringBase",
    "MonitoringEvent",
    "MonitoringLevel",
    "MonitoringLevelMapping",
    "MonitoringMetric",
    "MonitoringSystem",
    "OpsGenieLevel",
)


class MonitoringSystem(Enum):
    """List of monitoring systems."""

    DATADOG = "datadog"
    OPSGENIE = "opsgenie"
    LOGGING = "logging"


class DatadogLevel(Enum):
    """Datadog alert levels."""

    error = "error"
    warning = "warning"
    info = "info"
    success = "success"


class OpsGenieLevel(Enum):
    """OpsGenie alert levels."""

    P1 = "P1"
    P2 = "P2"
    P3 = "P3"
    P4 = "P4"
    P5 = "P5"


class MonitoringBase(GatewayStruct):
    """
    Monitoring base class.

    Attributes:
        tags (Dict[str, str]): Monitoring tags.
        tag_str (str): String representation of tags.
    """

    tags: Dict[str, str] = {}
    tag_str: str = ""

    def __init__(self, **kwargs):
        # We add override __init__ here since csp does not currently
        # provide type validation for dict typing so we
        # perform it here manually using csp's internal typing
        # validation
        super().__init__(**kwargs)
        obj_type = self.__class__
        # based on from csp
        for field, value in kwargs.items():
            expected_type = obj_type.__full_metadata_typed__.get(field, None)
            if expected_type is None:
                raise KeyError(f"Unexpected field `{field}` for type {obj_type}")
            setattr(self, field, obj_type._obj_from_python(value, expected_type))

    def get_tags(self, extra_tags: Dict[str, str] = None) -> List[str]:
        """
        Appends extra (i.e., global) tags if supplied to event or metric specific tags.
        Transforms dictionary of monitoring tags into Datadog compliant List format.
        i.e., {'foo':'bar', 'baz': 'qux'} transformed into ['foo:bar', 'baz:qux'].

        Args:
            extra_tags (Optional[Dict[str, str]]): Extra tags to be appended.

        Returns:
            List[str]: Datadog tags list.
        """
        if extra_tags:
            self.tags.update(extra_tags)
        if "monitoring_id" not in self.tags:
            self.tags["monitoring_id"] = self.id
        return [f"{k}:{v}" for k, v in self.tags.items()]

    def get_datadog_timestamp(self, from_tz: timezone = timezone.utc, to_tz: timezone = None) -> float:
        """
        Provides Datadog timestamp.
        Optionally converts timestamp time zone into Datadog instance time zone.
        This is needed as Datadog does not recognize future timestamps.

        Args:
            from_tz (timezone, optional): The original time zone of the timestamp. Defaults to timezone.utc.
            to_tz (Optional[timezone]): The target time zone for conversion. Defaults to None.

        Returns:
            float: Datadog timestamp.
        """
        return self.timestamp.replace(tzinfo=from_tz).astimezone(tz=to_tz).timestamp()

    @abstractmethod
    def to_datadog(extra_tags: List[str] = None) -> Dict[str, Union[str, List[str]]]:
        """
        Abstract method to implement conversion to Datadog payload.

        Args:
            extra_tags (Optional[List[str]]): Metric extra tags.

        Returns:
            Dict[str, Union[str, List[str]]]: Datadog metric payload.
        """
        raise NotImplementedError

    @abstractmethod
    def to_dict(self, extra_tags: Dict[str, str] = None) -> Dict[str, str]:
        """
        Abstract method to convert to dictionary.

        Args:
            extra_tags (Optional[Dict[str, str]]): Extra tags to be appended.

        Returns:
            Dict[str, str]: Dictionary representation.
        """
        raise NotImplementedError

    @abstractmethod
    def to_opsgenie(self, extra_tags: Dict[str, str] = None) -> "opsgenie_sdk.CloseAlertPayload":  # noqa: F821
        """
        Abstract method to convert to OpsGenie payload.

        Args:
            extra_tags (Optional[Dict[str, str]]): Extra tags to be appended.

        Returns:
            opsgenie_sdk.CloseAlertPayload: OpsGenie payload.
        """
        raise NotImplementedError

    @override
    def psp_flatten(self, custom_jsonifier=None):
        """
        Flattens the object for PSP.

        Args:
            custom_jsonifier (optional): Custom JSON serializer.

        Returns:
            The flattened object.
        """
        self.tag_str = orjson.dumps(self.tags).decode()
        return super().psp_flatten(custom_jsonifier)


class MonitoringEvent(MonitoringBase):
    """
    Monitoring event for Datadog.

    Attributes:
        title (str): Datadog event title.
        text (str): Datadog event text.
        event_type (str): Datadog key by which to group events in the event stream.
        alert_type (DatadogLevel): Datadog alert type, e.g., error, warning, etc.
        source (str): Source type of the event.
    """

    DEFAULT_SOURCE = "python"

    # leave for backward compatibility
    AlertType = DatadogLevel

    title: str
    text: str
    event_type: str
    alert_type: Union[AlertType, DatadogLevel] = DatadogLevel.info
    source: str = DEFAULT_SOURCE

    def get_event_alias(
        self,
        category_tag: str = "event_group",
        alias_tags: Dict[str, str] = None,
        extra_tags: Dict[str, str] = None,
        separator: str = ":",
    ) -> str:
        """
        Formats alias string from specified tags.
        This functionality is useful to aggregate OpsGenie events under the same alias.

        Args:
            category_tag (str): Name of the tag used to categorize the evnet. Default is "event_group".
            alias_tags (Optional[Dict[str, str]]): Event tags used to build an alias.
            extra_tags (Optional[Dict[str, str]]): Event extra tags.
            separator (str): Separator to build alias. Default is ":".
        Returns:
            (str): Aggregator alias
        """
        if alias_tags:
            if not extra_tags:
                extra_tags = {}
            tags = self.tags | extra_tags
            alias = []
            if category_tag in tags and tags[category_tag] in alias_tags:
                for alias_tag in alias_tags[tags[category_tag]]:
                    if alias_tag in tags:
                        alias.append(tags[alias_tag])
            if alias:
                return separator.join(alias)
        return ""

    def to_datadog(self, extra_tags: Dict[str, str] = None) -> Dict[str, Union[str, List[str]]]:
        """
        Formats Datadog event payload.

        Args:
            extra_tags (Optional[Dict[str, str]]): Metric extra tags.

        Returns:
            Dict[str, Union[str, List[str]]]: Datadog event payload.
        """
        return {
            "title": self.title,
            "text": self.text,
            "alert_type": self.alert_type.value,
            "aggregation_key": self.event_type,
            "date_happened": self.get_datadog_timestamp(),
            "source_type_name": self.source,
            "tags": self.get_tags(extra_tags),
        }

    def to_dict(self, extra_tags: Dict[str, str] = None) -> Dict[str, str]:
        """
        Converts the event to a dictionary.

        Args:
            extra_tags (Optional[Dict[str, str]]): Extra tags to be appended.

        Returns:
            Dict[str, str]: Dictionary representation of the event.
        """
        attrs = (
            "title",
            "text",
            "source",
            "alert_type",
            "event_type",
            "id",
            "timestamp",
        )
        attrs_dict = {attr: getattr(self, attr) for attr in attrs}
        attrs_dict["tags"] = self.get_tags(extra_tags=extra_tags)
        return attrs_dict

    def to_opsgenie(
        self,
        category_tag: str = "event_group",
        alias_tags: Dict[str, str] = None,
        extra_tags: Dict[str, str] = None,
        separator: str = ":",
    ) -> "opsgenie_sdk.CloseAlertPayload":  # noqa: F821
        """
        Converts the event to an OpsGenie payload.

        Args:
            category_tag (str): Name of the tag used to categorize the evnet. Default is "event_group".
            alias_tags (Optional[Dict[str, str]]): Event tags used to build an alias.
            extra_tags (Optional[Dict[str, str]]): Event extra tags.
            separator (str): Separator to build alias. Default is ":".
        Returns:
            opsgenie_sdk.CloseAlertPayload: OpsGenie payload.
        """
        import opsgenie_sdk

        if not extra_tags:
            extra_tags = {}
        alias = self.get_event_alias(
            category_tag=category_tag,
            alias_tags=alias_tags,
            extra_tags=extra_tags,
            separator=separator,
        )
        return opsgenie_sdk.CreateAlertPayload(
            source=self.source,
            message=self.title,
            alias=alias or self.id,
            description=self.text,
            tags=self.get_tags(extra_tags=extra_tags),
            entity=self.event_type,
            details=self.tags | extra_tags,
            priority=MonitoringLevelMapping.from_event(self).opsgenie,
        )


class MonitoringLevel(BaseModel):
    """Dataclass used for level mapping accross systems"""

    level: str
    datadog: DatadogLevel
    opsgenie: OpsGenieLevel
    logging: int


class MonitoringLevelMapping(Enum):
    """
    MonitoringLevelMapping provides mapping between Datadog, OpsGenie, and logging levels.

    Attributes:
        critical (MonitoringLevel): Critical level mapping.
        error (MonitoringLevel): Error level mapping.
        warning (MonitoringLevel): Warning level mapping.
        info (MonitoringLevel): Info level mapping.
        debug (MonitoringLevel): Debug level mapping.
    """

    critical = MonitoringLevel(
        level="critical",
        datadog=DatadogLevel.error,
        opsgenie=OpsGenieLevel.P1,
        logging=50,
    )
    error = MonitoringLevel(level="error", datadog=DatadogLevel.error, opsgenie=OpsGenieLevel.P2, logging=40)
    warning = MonitoringLevel(
        level="warning",
        datadog=DatadogLevel.warning,
        opsgenie=OpsGenieLevel.P3,
        logging=30,
    )
    info = MonitoringLevel(level="info", datadog=DatadogLevel.info, opsgenie=OpsGenieLevel.P4, logging=20)
    debug = MonitoringLevel(
        level="debug",
        datadog=DatadogLevel.success,
        opsgenie=OpsGenieLevel.P5,
        logging=10,
    )

    @property
    def opsgenie(self) -> str:
        """Gets the OpsGenie level value."""
        return self.value.opsgenie.value

    @property
    def datadog(self) -> str:
        """Gets the Datadog level value."""
        return self.value.datadog.value

    @property
    def logging(self) -> str:
        """Gets the logging level."""
        return self.value.logging

    @classmethod
    def from_event(cls, event: MonitoringEvent) -> "MonitoringLevelMapping":
        """
        Gets the monitoring level mapping from a MonitoringEvent.

        Args:
            event (MonitoringEvent): The monitoring event.

        Returns:
            MonitoringLevelMapping: The corresponding monitoring level mapping.

        Raises:
            ValueError: If no mapping is found for the event.
        """
        mapping: Dict[str, MonitoringLevelMapping] = {member.value.datadog: member for member in cls}
        if event.alert_type in mapping:
            return mapping[event.alert_type]
        raise ValueError(f"No MonitoringLevelMapping found for {event}")

    @classmethod
    def from_alert_type(cls, level: Union[int, str, DatadogLevel, OpsGenieLevel]) -> "MonitoringLevelMapping":
        """
        Gets the monitoring level mapping from an alert type.

        Args:
            level Union[int, str, DatadogLevel, OpsGenieLevel]: The alert level.

        Returns:
            MonitoringLevelMapping: The corresponding monitoring level mapping.

        Raises:
            ValueError: If no mapping is found for the level.DatadogLevel
        """
        if isinstance(level, str):
            try:
                level = DatadogLevel(level)
            except ValueError:
                try:
                    level = OpsGenieLevel(level)
                except ValueError:
                    raise ValueError(f"No alert level {level} found")
        if isinstance(level, DatadogLevel):
            # Datadog has no critical level
            if level == DatadogLevel.error:
                return cls.error
            for alert_level in cls:
                if getattr(alert_level, MonitoringSystem.DATADOG.value) == level.value:
                    return alert_level
        elif isinstance(level, OpsGenieLevel):
            for alert_level in cls:
                if getattr(alert_level, MonitoringSystem.OPSGENIE.value) == level.value:
                    return alert_level
        elif isinstance(level, int):
            for alert_level in cls:
                if getattr(alert_level, MonitoringSystem.LOGGING.value) == level:
                    return alert_level
        raise ValueError(f"No alert level {level} found ---")

    @classmethod
    def get_opsgenie_levels(cls) -> Tuple[str]:
        """
        Gets all OpsGenie levels.

        Returns:
            Tuple[str]: All OpsGenie levels.
        """
        return tuple(member.opsgenie for member in cls)

    @classmethod
    def get_datadog_levels(cls) -> Tuple[str]:
        """
        Gets all Datadog levels.

        Returns:
            Tuple[str]: All Datadog levels.
        """
        return tuple(member.datadog for member in cls)

    @classmethod
    def get_logging_levels(cls) -> Tuple[str]:
        """
        Gets all logging levels.

        Returns:
            Tuple[str]: All logging levels.
        """
        return tuple(member.logging for member in cls)


class MonitoringMetric(MonitoringBase):
    """
    Monitoring metric for Datadog.

    Attributes:
        metric (str): Datadog metric name.
        metric_type (MetricType): Datadog metric type, i.e., count, rate, gauge.
        value (Union[int, float]): Numeric metric value.
    """

    class MetricType(Enum):
        """
        Metric type struct
        """

        gauge = "gauge"
        count = "count"
        rate = "rate"

    metric: str
    metric_type: MetricType
    value: float

    def to_datadog(self, extra_tags: Dict[str, str] = None) -> Dict[str, Union[str, List[str]]]:
        """
        Formats Datadog metric payload.

        Args:
            extra_tags (Optional[Dict[str, str]]): Metric extra tags.

        Returns:
            Dict[str, Union[str, List[str]]]: Datadog metric payload.
        """
        return {
            "metric": self.metric,
            "type": self.metric_type.value,
            "points": [(self.get_datadog_timestamp(), self.value)],
            "tags": self.get_tags(extra_tags),
        }

    def to_dict(self, extra_tags: Dict[str, str] = None) -> Dict[str, Union[str, List[str]]]:
        """
        Converts the metric to a dictionary.

        Args:
            extra_tags (Optional[Dict[str, str]]): Extra tags to be appended.

        Returns:
            Dict[str, Union[str, List[str]]]: Dictionary representation of the metric.
        """
        return self.to_datadog(extra_tags=extra_tags)

    def to_opsgenie(
        self,
        title: str,
        alert_type: MonitoringLevelMapping,
        text: str = None,
        source: str = None,
        extra_tags: Dict[str, str] = None,
    ) -> "opsgenie_sdk.CloseAlertPayload":  # noqa: F821
        """
        Converts the metric to an OpsGenie payload.

        Args:
            title (str): Alert title.
            alert_type (MonitoringLevelMapping): Alert type mapping.
            text (Optional[str]): Alert text. Defaults to None.
            source (Optional[str]): Source of the alert. Defaults to None.
            extra_tags (Optional[Dict[str, str]]): Extra tags to be appended. Defaults to None.

        Returns:
            opsgenie_sdk.CloseAlertPayload: OpsGenie payload.
        """
        import opsgenie_sdk

        if not extra_tags:
            extra_tags = {}
        return opsgenie_sdk.CreateAlertPayload(
            source=source or MonitoringEvent.DEFAULT_SOURCE,
            message=title,
            alias=self.id,
            description=text or title,
            tags=self.get_tags(extra_tags=extra_tags),
            entity=self.metric,
            details=self.tags | extra_tags,
            priority=alert_type.opsgenie,
            note=text or title,
        )



from collections import defaultdict
from datetime import datetime, timedelta, timezone
from typing import Any, List, Union
from unittest import mock

import csp
import numpy as np
import orjson
import pytest
from csp import ts
from csp.adapters.kafka import KafkaStartOffset as KafkaStartOffsetCsp
from csp.typing import Numpy1DArray
from pydantic import ValidationError

from csp_gateway import (
    AddChannelsToGraphOutput,
    ChannelSelection,
    EncodedEngineCycle,
    GatewayStruct,
    KafkaChannelProcessor,
    KafkaConfiguration,
    ReadWriteKafka,
    ReadWriteMode,
    ReplayEngineKafka,
)
from csp_gateway.testing.shared_helpful_classes import (
    MyGateway,
    MyGatewayChannels,
    MySetModule,
    MyStruct,
)


class MyNumpyStruct(MyStruct):
    arr: Numpy1DArray[float]


class MyGatewayChannelsWithNumpyStruct(MyGatewayChannels):
    my_numpy_struct_channel: ts[MyNumpyStruct] = None


class ProcessMyFlagTrue(KafkaChannelProcessor):
    def process(self, obj: Union[List[GatewayStruct], GatewayStruct], topic: str, key: str) -> Any:
        if isinstance(obj, list):
            return obj if all(getattr(val, "my_flag", False) is True for val in obj) else None
        return obj if getattr(obj, "my_flag", False) is True else None


class FloatKafkaChannelProcessor(KafkaChannelProcessor):
    goal_float: float
    key: str

    def process(self, obj: Union[List[GatewayStruct], GatewayStruct], topic: str, key: str) -> bool:
        if key == self.key:
            obj.foo = self.goal_float
            return obj
        return obj


def create_kafka_engine_replay(selection: ChannelSelection, read_write_mode: ReadWriteMode):
    broker = "kafka-broker:9093"
    topic = "topic_test"
    key = "test_kafka_gateway_module"
    config = KafkaConfiguration(broker=broker)
    return ReplayEngineKafka(
        selection=selection,
        config=config,
        topic=topic,
        key=key,
        read_write_mode=read_write_mode,
    )


def test_config_validation_fails():
    broker = "kafka-broker:9093"
    with pytest.raises(ValidationError):
        KafkaConfiguration(broker=broker, sasl_kerberos_keytab="Fake/location")


def test_config_pydantic_validation():
    broker = "kafka-broker:9093"
    for offset_name, val in KafkaStartOffsetCsp.__members__.items():
        config = KafkaConfiguration(broker=broker, start_offset=offset_name)
        assert config.start_offset.name == offset_name
        config = KafkaConfiguration(broker=broker, start_offset=val)
        assert config.start_offset.name == offset_name


def test_subscribe_with_csp_engine_timestamp_only_if_encoding_expected():
    config = KafkaConfiguration(broker="dummy")
    ReadWriteKafka(config=config, encoding_with_engine_timestamps=False, subscribe_with_csp_engine_timestamp=False)
    ReadWriteKafka(config=config, encoding_with_engine_timestamps=True, subscribe_with_csp_engine_timestamp=False)
    ReadWriteKafka(config=config, encoding_with_engine_timestamps=True, subscribe_with_csp_engine_timestamp=False)
    with pytest.raises(ValidationError):
        ReadWriteKafka(config=config, encoding_with_engine_timestamps=False, subscribe_with_csp_engine_timestamp=True)


@mock.patch("csp.adapters.kafka.KafkaAdapterManager", autospec=True)
@pytest.mark.parametrize("by_key", [True, False])
def test_kafka_engine_replay_write(mock_object, by_key):
    mock_publish = mock.MagicMock()
    mock_subscribe = mock.MagicMock()
    mock_object.return_value.subscribe = mock_subscribe
    mock_object.return_value.publish = mock_publish

    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    channels = MyGatewayChannels.fields()
    kafka_engine_replay_write = create_kafka_engine_replay(ChannelSelection.model_validate(channels), ReadWriteMode.WRITE)

    gateway_writing = MyGateway(
        modules=[setter, kafka_engine_replay_write, AddChannelsToGraphOutput()],
        channels=MyGatewayChannels(),
    )
    out = csp.run(gateway_writing.graph, starttime=datetime.now(timezone.utc), endtime=timedelta(seconds=5))
    assert mock_publish.call_count == 1
    assert mock_subscribe.call_count == 0

    # Does not interfere with graph
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_array_channel"]) == 1
    assert len(out["my_array_channel"][0][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1


@mock.patch("csp.adapters.kafka.KafkaAdapterManager", autospec=True)
@pytest.mark.parametrize("read_write_mode", [ReadWriteMode.READ, ReadWriteMode.READ_AND_WRITE])
def test_kafka_engine_replay_read_and_write_read(mock_object, read_write_mode):
    mock_publish = mock.MagicMock()

    def mock_subscribe(*a, **kw):
        return csp.null_ts(EncodedEngineCycle)

    mock_object.return_value.subscribe.side_effect = mock_subscribe
    mock_object.return_value.publish = mock_publish

    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    kafka_engine_replay_write = create_kafka_engine_replay(ChannelSelection.model_validate(channels), read_write_mode)

    gateway_writing = MyGateway(
        modules=[setter, kafka_engine_replay_write, setter],
        channels=MyGatewayChannels(),
    )
    csp.run(gateway_writing.graph, starttime=datetime.now(timezone.utc), endtime=timedelta(seconds=5))
    if read_write_mode is ReadWriteMode.READ_AND_WRITE:
        assert mock_publish.call_count == 1
    else:
        assert mock_publish.call_count == 0


@mock.patch("csp.adapters.kafka.KafkaAdapterManager", autospec=True)
def test_kafka_read_write_fails_with_dict_basket(mock_object):
    mock_publish = mock.MagicMock()

    def mock_subscribe(*a, **kw):
        return csp.null_ts(EncodedEngineCycle)

    mock_object.return_value.subscribe.side_effect = mock_subscribe
    mock_object.return_value.publish = mock_publish

    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    broker = "kafka-broker:9093"
    config = KafkaConfiguration(broker=broker)

    kafka_writer = ReadWriteKafka(
        config=config,
        publish_channel_to_topic_and_key={
            MyGatewayChannels.my_channel: {"topic1": "key"},
            MyGatewayChannels.my_str_basket: {"topic1": "key2"},
            MyGatewayChannels.my_enum_basket: {"topic2": "key"},
        },
    )

    gateway_writing = MyGateway(
        modules=[setter, kafka_writer, setter],
        channels=MyGatewayChannels(),
    )
    with pytest.raises((TypeError, AttributeError)):
        csp.run(
            gateway_writing.graph,
            starttime=datetime.now(timezone.utc),
            endtime=timedelta(seconds=5),
        )


@mock.patch("csp.adapters.kafka.KafkaAdapterManager", autospec=True)
@pytest.mark.parametrize("encoding_with_engine_timestamps", [True, False])
def test_kafka_read_write_read(mock_object, encoding_with_engine_timestamps, capsys):
    def mock_publish_action(*a, **kw):
        if kw["topic"] == "kafka_topic1" and kw["key"] == "kafka_key2":
            # If Kafka publishes to this key on this topic, we capture it
            # with a print statement.
            if encoding_with_engine_timestamps:
                csp.print("val_check", kw["x"].encoding)
            else:
                csp.print("val_check", kw["x"])

    mock_publish = mock.MagicMock(side_effect=mock_publish_action)
    # mock_publish = mock.MagicMock()

    def mock_subscribe(*a, **kw):
        return csp.null_ts(EncodedEngineCycle)

    mock_object.return_value.subscribe.side_effect = mock_subscribe
    mock_object.return_value.publish = mock_publish

    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    broker = "kafka-broker:9093"
    config = KafkaConfiguration(broker=broker)

    kafka_reader = ReadWriteKafka(
        config=config,
        publish_channel_to_topic_and_key={
            MyGatewayChannelsWithNumpyStruct.my_channel: {"kafka_topic1": "kafka_key"},
            MyGatewayChannelsWithNumpyStruct.my_array_channel: {"kafka_topic1": "kafka_key2"},
            MyGatewayChannelsWithNumpyStruct.my_numpy_struct_channel: {"kafka_topic2": "kafka_key"},
        },
        encoding_with_engine_timestamps=encoding_with_engine_timestamps,
    )

    gateway_writing = MyGateway(
        channels_model=MyGatewayChannelsWithNumpyStruct,
        modules=[setter, kafka_reader, setter],
        channels=MyGatewayChannelsWithNumpyStruct(),
    )
    csp.run(gateway_writing.graph, starttime=datetime.now(timezone.utc), endtime=timedelta(seconds=5))
    calls = mock_publish.call_args_list
    topic_and_call_args = defaultdict(lambda *a, **kw: defaultdict(int))
    for call in calls:
        kwargs = call.kwargs
        key = kwargs["key"]
        topic = kwargs["topic"]
        topic_and_call_args["key"][key] += 1
        topic_and_call_args["topic"][topic] += 1

    assert topic_and_call_args["key"]["kafka_key"] == 2
    assert topic_and_call_args["key"]["kafka_key2"] == 1
    assert topic_and_call_args["topic"]["kafka_topic1"] == 2
    assert topic_and_call_args["topic"]["kafka_topic2"] == 1
    captured = capsys.readouterr()
    assert len(captured.out) > 0


@mock.patch("csp.adapters.kafka.KafkaAdapterManager", autospec=True)
def test_kafka_read_write_write(mock_object):
    mock_publish = mock.MagicMock()

    def mock_subscribe(*a, **kw):
        if kw["topic"] == "kafka_topic1" and kw["key"] == "kafka_key":
            encoding = orjson.dumps({"foo": 9.1}).decode()
            return csp.const(EncodedEngineCycle(encoding=encoding))
        elif kw["topic"] == "kafka_topic1" and kw["key"] == "kafka_key2":
            encoding = orjson.dumps([{"foo": 0.1}, {"my_flag": False}]).decode()
            return csp.const(EncodedEngineCycle(encoding=encoding), timedelta(seconds=1))
        elif kw["topic"] == "kafka_topic2" and kw["key"] == "kafka_key":
            encoding = orjson.dumps({"arr": [7.2, -1001.3]}).decode()
            return csp.const(EncodedEngineCycle(encoding=encoding), timedelta(seconds=1))
        else:
            raise ValueError("This should never be hit")

    mock_object.return_value.subscribe.side_effect = mock_subscribe
    mock_object.return_value.publish = mock_publish

    broker = "kafka-broker:9093"
    config = KafkaConfiguration(broker=broker)

    kafka_writer = ReadWriteKafka(
        config=config,
        subscribe_channel_to_topic_and_key={
            MyGatewayChannelsWithNumpyStruct.my_channel: {"kafka_topic1": "kafka_key"},
            MyGatewayChannelsWithNumpyStruct.my_list_channel: {"kafka_topic1": "kafka_key2"},
            MyGatewayChannelsWithNumpyStruct.my_numpy_struct_channel: {"kafka_topic2": "kafka_key"},
        },
    )

    gateway_writing = MyGateway(
        channels_model=MyGatewayChannelsWithNumpyStruct,
        modules=[kafka_writer, AddChannelsToGraphOutput()],
        channels=MyGatewayChannelsWithNumpyStruct(),
    )
    out = csp.run(gateway_writing.graph, starttime=datetime.now(timezone.utc), endtime=timedelta(seconds=5))
    assert not mock_publish.call_args_list
    assert len(out[MyGatewayChannelsWithNumpyStruct.my_channel]) == 1
    assert out[MyGatewayChannelsWithNumpyStruct.my_channel][0][1].foo == 9.1

    assert len(out[MyGatewayChannelsWithNumpyStruct.my_list_channel]) == 1
    assert len(out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1]) == 2
    assert out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1][0].foo == 0.1
    assert out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1][1].my_flag is False

    assert len(out[MyGatewayChannelsWithNumpyStruct.my_numpy_struct_channel]) == 1
    my_numpy_struct = out[MyGatewayChannelsWithNumpyStruct.my_numpy_struct_channel][0][1]
    np.testing.assert_array_almost_equal(my_numpy_struct.arr, np.array([7.2, -1001.3]))


@mock.patch("csp.adapters.kafka.KafkaAdapterManager", autospec=True)
@pytest.mark.parametrize("use_struct_id", [True, False])
@pytest.mark.parametrize("use_struct_timestamp", [True, False])
def test_kafka_read_write_write_overwrite_id_timestamp(mock_object, use_struct_id, use_struct_timestamp):
    mock_publish = mock.MagicMock()

    def mock_subscribe(*a, **kw):
        if kw["topic"] == "kafka_topic1" and kw["key"] == "kafka_key":
            encoding = orjson.dumps({"foo": 9.1, "id": "howdy"}).decode()
            return csp.const(EncodedEngineCycle(encoding=encoding))
        elif kw["topic"] == "kafka_topic1" and kw["key"] == "kafka_key2":
            encoding = orjson.dumps([{"foo": 0.1, "timestamp": datetime(2021, 1, 1)}, {"my_flag": False}]).decode()
            return csp.const(EncodedEngineCycle(encoding=encoding), timedelta(seconds=1))
        else:
            raise ValueError("This should never be hit")

    mock_object.return_value.subscribe.side_effect = mock_subscribe
    mock_object.return_value.publish = mock_publish

    broker = "kafka-broker:9093"
    config = KafkaConfiguration(broker=broker)

    kafka_writer = ReadWriteKafka(
        config=config,
        subscribe_with_struct_id=use_struct_id,
        subscribe_with_struct_timestamp=use_struct_timestamp,
        subscribe_channel_to_topic_and_key={
            MyGatewayChannelsWithNumpyStruct.my_channel: {"kafka_topic1": "kafka_key"},
            MyGatewayChannelsWithNumpyStruct.my_list_channel: {"kafka_topic1": "kafka_key2"},
        },
    )

    gateway_writing = MyGateway(
        channels_model=MyGatewayChannelsWithNumpyStruct,
        modules=[kafka_writer, AddChannelsToGraphOutput()],
        channels=MyGatewayChannelsWithNumpyStruct(),
    )
    out = csp.run(gateway_writing.graph, starttime=datetime.now(timezone.utc), endtime=timedelta(seconds=5))
    assert not mock_publish.call_args_list
    assert len(out[MyGatewayChannelsWithNumpyStruct.my_channel]) == 1
    assert out[MyGatewayChannelsWithNumpyStruct.my_channel][0][1].foo == 9.1
    if use_struct_id:
        assert out[MyGatewayChannelsWithNumpyStruct.my_channel][0][1].id == "howdy"
    else:
        assert out[MyGatewayChannelsWithNumpyStruct.my_channel][0][1].id != "howdy"

    assert len(out[MyGatewayChannelsWithNumpyStruct.my_list_channel]) == 1
    assert len(out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1]) == 2
    assert out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1][0].foo == 0.1
    if use_struct_timestamp:
        assert out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1][0].timestamp == datetime(2021, 1, 1)
    else:
        # this is in the future, so we know the datetime is past the current date
        assert out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1][0].timestamp > datetime(2023, 1, 1)
    assert out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1][1].my_flag is False


@mock.patch("csp.adapters.kafka.KafkaAdapterManager", autospec=True)
@pytest.mark.parametrize("my_flag", [True, False])
def test_kafka_read_write_filter_write(mock_object, my_flag):
    mock_publish = mock.MagicMock()

    def mock_subscribe(*a, **kw):
        if kw["topic"] == "kafka_topic1" and kw["key"] == "kafka_key":
            encoding = orjson.dumps({"foo": 9.1}).decode()
            return csp.const(EncodedEngineCycle(encoding=encoding))
        elif kw["topic"] == "kafka_topic1" and kw["key"] == "kafka_key2":
            encoding = orjson.dumps([{"foo": 0.1}, {"my_flag": my_flag}]).decode()
            return csp.const(EncodedEngineCycle(encoding=encoding), timedelta(seconds=1))
        elif kw["topic"] == "kafka_topic2" and kw["key"] == "kafka_key":
            encoding = orjson.dumps({"arr": [7.2, -1001.3]}).decode()
            return csp.const(EncodedEngineCycle(encoding=encoding), timedelta(seconds=1))
        else:
            raise ValueError("This should never be hit")

    mock_object.return_value.subscribe.side_effect = mock_subscribe
    mock_object.return_value.publish = mock_publish

    broker = "kafka-broker:9093"
    config = KafkaConfiguration(broker=broker)

    my_list_channel_processor = ProcessMyFlagTrue()
    kafka_writer = ReadWriteKafka(
        config=config,
        subscribe_channel_to_topic_and_key={
            MyGatewayChannelsWithNumpyStruct.my_channel: {"kafka_topic1": "kafka_key"},
            MyGatewayChannelsWithNumpyStruct.my_list_channel: {"kafka_topic1": "kafka_key2"},
            MyGatewayChannelsWithNumpyStruct.my_numpy_struct_channel: {"kafka_topic2": "kafka_key"},
        },
        subscribe_channel_processors={MyGatewayChannelsWithNumpyStruct.my_list_channel: my_list_channel_processor},
    )

    gateway_writing = MyGateway(
        channels_model=MyGatewayChannelsWithNumpyStruct,
        modules=[kafka_writer, AddChannelsToGraphOutput()],
        channels=MyGatewayChannelsWithNumpyStruct(),
    )
    out = csp.run(gateway_writing.graph, starttime=datetime.now(timezone.utc), endtime=timedelta(seconds=5))
    assert not mock_publish.call_args_list
    assert len(out[MyGatewayChannelsWithNumpyStruct.my_channel]) == 1
    assert out[MyGatewayChannelsWithNumpyStruct.my_channel][0][1].foo == 9.1

    # in `my_list_channel` structs get filtered out if `my_flag` is set to
    # False.
    if not my_flag:
        assert len(out[MyGatewayChannelsWithNumpyStruct.my_list_channel]) == 0
    else:
        assert len(out[MyGatewayChannelsWithNumpyStruct.my_list_channel]) == 1
        assert len(out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1]) == 2
        assert out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1][0].foo == 0.1
        assert out[MyGatewayChannelsWithNumpyStruct.my_list_channel][0][1][1].my_flag is True

    assert len(out[MyGatewayChannelsWithNumpyStruct.my_numpy_struct_channel]) == 1
    my_numpy_struct = out[MyGatewayChannelsWithNumpyStruct.my_numpy_struct_channel][0][1]
    np.testing.assert_array_almost_equal(my_numpy_struct.arr, np.array([7.2, -1001.3]))


@mock.patch("csp.adapters.kafka.KafkaAdapterManager", autospec=True)
@pytest.mark.parametrize("encoding_with_engine_timestamps", [True, False])
@pytest.mark.parametrize("filter_key", ["kafka_key", "non_existent_key"])
def test_kafka_read_write_filter_read(mock_object, encoding_with_engine_timestamps, filter_key, capsys):
    def mock_publish_action(*a, **kw):
        if kw["topic"] == "kafka_topic1" and kw["key"] == "kafka_key":
            # If Kafka publishes to this key on this topic, we capture it
            # with a print statement.
            if encoding_with_engine_timestamps:
                csp.print("val_check", kw["x"].encoding)
            else:
                csp.print("val_check", kw["x"])

    mock_publish = mock.MagicMock(side_effect=mock_publish_action)

    def mock_subscribe(*a, **kw):
        return csp.null_ts(EncodedEngineCycle)

    mock_object.return_value.subscribe.side_effect = mock_subscribe
    mock_object.return_value.publish = mock_publish

    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=99.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    broker = "kafka-broker:9093"
    config = KafkaConfiguration(broker=broker)

    my_channel_processor = FloatKafkaChannelProcessor(goal_float=1.0, key=filter_key)

    kafka_writer = ReadWriteKafka(
        config=config,
        publish_channel_to_topic_and_key={
            MyGatewayChannelsWithNumpyStruct.my_channel: {"kafka_topic1": "kafka_key"},
            MyGatewayChannelsWithNumpyStruct.my_array_channel: {"kafka_topic1": "kafka_key2"},
            MyGatewayChannelsWithNumpyStruct.my_numpy_struct_channel: {"kafka_topic2": "kafka_key"},
        },
        publish_channel_processors={MyGatewayChannelsWithNumpyStruct.my_channel: my_channel_processor},
        encoding_with_engine_timestamps=encoding_with_engine_timestamps,
    )

    gateway_writing = MyGateway(
        channels_model=MyGatewayChannelsWithNumpyStruct,
        modules=[setter, kafka_writer, setter],
        channels=MyGatewayChannelsWithNumpyStruct(),
    )
    csp.run(gateway_writing.graph, starttime=datetime.now(timezone.utc), endtime=timedelta(seconds=5))
    calls = mock_publish.call_args_list
    topic_and_call_args = defaultdict(lambda *a, **kw: defaultdict(int))
    for call in calls:
        kwargs = call.kwargs
        key = kwargs["key"]
        topic = kwargs["topic"]
        topic_and_call_args["key"][key] += 1
        topic_and_call_args["topic"][topic] += 1

    assert topic_and_call_args["key"]["kafka_key"] == 2
    assert topic_and_call_args["key"]["kafka_key2"] == 1
    assert topic_and_call_args["topic"]["kafka_topic1"] == 2
    assert topic_and_call_args["topic"]["kafka_topic2"] == 1
    captured = capsys.readouterr()
    if filter_key == "kafka_key":
        assert captured.out.count('"foo":1.0') == 2
        assert captured.out.count('"foo":99.0') == 0
    else:
        assert captured.out.count('"foo":1.0') == 0
        assert captured.out.count('"foo":99.0') == 2



import logging
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

from ccflow import RootModelRegistry, load_config as load_config_base

if TYPE_CHECKING:
    from csp_gateway import Gateway

log = logging.getLogger(__name__)

__all__ = (
    "load_config",
    "load_gateway",
)


def load_config(
    config_dir: str = "",
    config_name: str = "",
    overrides: Optional[List[str]] = None,
    *,
    overwrite: bool = True,
    basepath: str = "",
) -> RootModelRegistry:
    log.info("Loading csp-gateway config...")
    return load_config_base(
        root_config_dir=str(Path(__file__).resolve().parent),
        root_config_name="base",
        config_dir=config_dir,
        config_name=config_name,
        overrides=overrides,
        overwrite=overwrite,
        basepath=basepath,
    )


@wraps(load_config)
def load_gateway(*args, **kwargs) -> "Gateway":
    log.info("Loading csp-gateway config...")
    return load_config(*args, **kwargs)["gateway"]
